from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)
from ._partial_bid_matrix_calculation_input import (
    PartialBidMatrixCalculationInput,
    PartialBidMatrixCalculationInputWrite,
)

if TYPE_CHECKING:
    from ._bid_configuration_day_ahead import (
        BidConfigurationDayAhead,
        BidConfigurationDayAheadGraphQL,
        BidConfigurationDayAheadWrite,
    )
    from ._price_production import PriceProduction, PriceProductionGraphQL, PriceProductionWrite
    from ._shop_based_partial_bid_configuration import (
        ShopBasedPartialBidConfiguration,
        ShopBasedPartialBidConfigurationGraphQL,
        ShopBasedPartialBidConfigurationWrite,
    )


__all__ = [
    "MultiScenarioPartialBidMatrixCalculationInput",
    "MultiScenarioPartialBidMatrixCalculationInputWrite",
    "MultiScenarioPartialBidMatrixCalculationInputApply",
    "MultiScenarioPartialBidMatrixCalculationInputList",
    "MultiScenarioPartialBidMatrixCalculationInputWriteList",
    "MultiScenarioPartialBidMatrixCalculationInputApplyList",
    "MultiScenarioPartialBidMatrixCalculationInputFields",
    "MultiScenarioPartialBidMatrixCalculationInputTextFields",
]


MultiScenarioPartialBidMatrixCalculationInputTextFields = Literal[
    "workflow_execution_id", "function_name", "function_call_id"
]
MultiScenarioPartialBidMatrixCalculationInputFields = Literal[
    "workflow_execution_id", "workflow_step", "function_name", "function_call_id", "bid_date"
]

_MULTISCENARIOPARTIALBIDMATRIXCALCULATIONINPUT_PROPERTIES_BY_FIELD = {
    "workflow_execution_id": "workflowExecutionId",
    "workflow_step": "workflowStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
    "bid_date": "bidDate",
}


class MultiScenarioPartialBidMatrixCalculationInputGraphQL(GraphQLCore):
    """This represents the reading version of multi scenario partial bid matrix calculation input, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the multi scenario partial bid matrix calculation input.
        data_record: The data record of the multi scenario partial bid matrix calculation input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_date: The bid date
        bid_configuration: TODO description
        partial_bid_configuration: The partial bid configuration related to the bid calculation task
        price_production: An array of shop results with price/prod time series pairs for all plants included in the respective shop scenario
    """

    view_id = dm.ViewId("power_ops_core", "MultiScenarioPartialBidMatrixCalculationInput", "1")
    workflow_execution_id: Optional[str] = Field(None, alias="workflowExecutionId")
    workflow_step: Optional[int] = Field(None, alias="workflowStep")
    function_name: Optional[str] = Field(None, alias="functionName")
    function_call_id: Optional[str] = Field(None, alias="functionCallId")
    bid_date: Optional[datetime.date] = Field(None, alias="bidDate")
    bid_configuration: Optional[BidConfigurationDayAheadGraphQL] = Field(
        default=None, repr=False, alias="bidConfiguration"
    )
    partial_bid_configuration: Optional[ShopBasedPartialBidConfigurationGraphQL] = Field(
        default=None, repr=False, alias="partialBidConfiguration"
    )
    price_production: Optional[list[PriceProductionGraphQL]] = Field(default=None, repr=False, alias="priceProduction")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("bid_configuration", "partial_bid_configuration", "price_production", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> MultiScenarioPartialBidMatrixCalculationInput:
        """Convert this GraphQL format of multi scenario partial bid matrix calculation input to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return MultiScenarioPartialBidMatrixCalculationInput(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            bid_date=self.bid_date,
            bid_configuration=(
                self.bid_configuration.as_read()
                if isinstance(self.bid_configuration, GraphQLCore)
                else self.bid_configuration
            ),
            partial_bid_configuration=(
                self.partial_bid_configuration.as_read()
                if isinstance(self.partial_bid_configuration, GraphQLCore)
                else self.partial_bid_configuration
            ),
            price_production=[price_production.as_read() for price_production in self.price_production or []],
        )

    def as_write(self) -> MultiScenarioPartialBidMatrixCalculationInputWrite:
        """Convert this GraphQL format of multi scenario partial bid matrix calculation input to the writing format."""
        return MultiScenarioPartialBidMatrixCalculationInputWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            bid_date=self.bid_date,
            bid_configuration=(
                self.bid_configuration.as_write()
                if isinstance(self.bid_configuration, GraphQLCore)
                else self.bid_configuration
            ),
            partial_bid_configuration=(
                self.partial_bid_configuration.as_write()
                if isinstance(self.partial_bid_configuration, GraphQLCore)
                else self.partial_bid_configuration
            ),
            price_production=[price_production.as_write() for price_production in self.price_production or []],
        )


class MultiScenarioPartialBidMatrixCalculationInput(PartialBidMatrixCalculationInput):
    """This represents the reading version of multi scenario partial bid matrix calculation input.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the multi scenario partial bid matrix calculation input.
        data_record: The data record of the multi scenario partial bid matrix calculation input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_date: The bid date
        bid_configuration: TODO description
        partial_bid_configuration: The partial bid configuration related to the bid calculation task
        price_production: An array of shop results with price/prod time series pairs for all plants included in the respective shop scenario
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "MultiScenarioPartialBidMatrixCalculationInput"
    )
    price_production: Union[list[PriceProduction], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="priceProduction"
    )

    def as_write(self) -> MultiScenarioPartialBidMatrixCalculationInputWrite:
        """Convert this read version of multi scenario partial bid matrix calculation input to the writing version."""
        return MultiScenarioPartialBidMatrixCalculationInputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            workflow_execution_id=self.workflow_execution_id,
            workflow_step=self.workflow_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            bid_date=self.bid_date,
            bid_configuration=(
                self.bid_configuration.as_write()
                if isinstance(self.bid_configuration, DomainModel)
                else self.bid_configuration
            ),
            partial_bid_configuration=(
                self.partial_bid_configuration.as_write()
                if isinstance(self.partial_bid_configuration, DomainModel)
                else self.partial_bid_configuration
            ),
            price_production=[
                price_production.as_write() if isinstance(price_production, DomainModel) else price_production
                for price_production in self.price_production or []
            ],
        )

    def as_apply(self) -> MultiScenarioPartialBidMatrixCalculationInputWrite:
        """Convert this read version of multi scenario partial bid matrix calculation input to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MultiScenarioPartialBidMatrixCalculationInputWrite(PartialBidMatrixCalculationInputWrite):
    """This represents the writing version of multi scenario partial bid matrix calculation input.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the multi scenario partial bid matrix calculation input.
        data_record: The data record of the multi scenario partial bid matrix calculation input node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        bid_date: The bid date
        bid_configuration: TODO description
        partial_bid_configuration: The partial bid configuration related to the bid calculation task
        price_production: An array of shop results with price/prod time series pairs for all plants included in the respective shop scenario
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "MultiScenarioPartialBidMatrixCalculationInput"
    )
    price_production: Union[list[PriceProductionWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="priceProduction"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            MultiScenarioPartialBidMatrixCalculationInput,
            dm.ViewId("power_ops_core", "MultiScenarioPartialBidMatrixCalculationInput", "1"),
        )

        properties: dict[str, Any] = {}

        if self.workflow_execution_id is not None:
            properties["workflowExecutionId"] = self.workflow_execution_id

        if self.workflow_step is not None:
            properties["workflowStep"] = self.workflow_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.bid_date is not None or write_none:
            properties["bidDate"] = self.bid_date.isoformat() if self.bid_date else None

        if self.bid_configuration is not None:
            properties["bidConfiguration"] = {
                "space": self.space if isinstance(self.bid_configuration, str) else self.bid_configuration.space,
                "externalId": (
                    self.bid_configuration
                    if isinstance(self.bid_configuration, str)
                    else self.bid_configuration.external_id
                ),
            }

        if self.partial_bid_configuration is not None:
            properties["partialBidConfiguration"] = {
                "space": (
                    self.space
                    if isinstance(self.partial_bid_configuration, str)
                    else self.partial_bid_configuration.space
                ),
                "externalId": (
                    self.partial_bid_configuration
                    if isinstance(self.partial_bid_configuration, str)
                    else self.partial_bid_configuration.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_types", "PriceProduction")
        for price_production in self.price_production or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=price_production,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.bid_configuration, DomainModelWrite):
            other_resources = self.bid_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.partial_bid_configuration, DomainModelWrite):
            other_resources = self.partial_bid_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class MultiScenarioPartialBidMatrixCalculationInputApply(MultiScenarioPartialBidMatrixCalculationInputWrite):
    def __new__(cls, *args, **kwargs) -> MultiScenarioPartialBidMatrixCalculationInputApply:
        warnings.warn(
            "MultiScenarioPartialBidMatrixCalculationInputApply is deprecated and will be removed in v1.0. Use MultiScenarioPartialBidMatrixCalculationInputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "MultiScenarioPartialBidMatrixCalculationInput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class MultiScenarioPartialBidMatrixCalculationInputList(DomainModelList[MultiScenarioPartialBidMatrixCalculationInput]):
    """List of multi scenario partial bid matrix calculation inputs in the read version."""

    _INSTANCE = MultiScenarioPartialBidMatrixCalculationInput

    def as_write(self) -> MultiScenarioPartialBidMatrixCalculationInputWriteList:
        """Convert these read versions of multi scenario partial bid matrix calculation input to the writing versions."""
        return MultiScenarioPartialBidMatrixCalculationInputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> MultiScenarioPartialBidMatrixCalculationInputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MultiScenarioPartialBidMatrixCalculationInputWriteList(
    DomainModelWriteList[MultiScenarioPartialBidMatrixCalculationInputWrite]
):
    """List of multi scenario partial bid matrix calculation inputs in the writing version."""

    _INSTANCE = MultiScenarioPartialBidMatrixCalculationInputWrite


class MultiScenarioPartialBidMatrixCalculationInputApplyList(
    MultiScenarioPartialBidMatrixCalculationInputWriteList
): ...


def _create_multi_scenario_partial_bid_matrix_calculation_input_filter(
    view_id: dm.ViewId,
    workflow_execution_id: str | list[str] | None = None,
    workflow_execution_id_prefix: str | None = None,
    min_workflow_step: int | None = None,
    max_workflow_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    min_bid_date: datetime.date | None = None,
    max_bid_date: datetime.date | None = None,
    bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    partial_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(workflow_execution_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id))
    if workflow_execution_id and isinstance(workflow_execution_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("workflowExecutionId"), values=workflow_execution_id))
    if workflow_execution_id_prefix is not None:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id_prefix)
        )
    if min_workflow_step is not None or max_workflow_step is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("workflowStep"), gte=min_workflow_step, lte=max_workflow_step)
        )
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if min_bid_date is not None or max_bid_date is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("bidDate"),
                gte=min_bid_date.isoformat() if min_bid_date else None,
                lte=max_bid_date.isoformat() if max_bid_date else None,
            )
        )
    if bid_configuration and isinstance(bid_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": bid_configuration},
            )
        )
    if bid_configuration and isinstance(bid_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidConfiguration"),
                value={"space": bid_configuration[0], "externalId": bid_configuration[1]},
            )
        )
    if bid_configuration and isinstance(bid_configuration, list) and isinstance(bid_configuration[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in bid_configuration],
            )
        )
    if bid_configuration and isinstance(bid_configuration, list) and isinstance(bid_configuration[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in bid_configuration],
            )
        )
    if partial_bid_configuration and isinstance(partial_bid_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("partialBidConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": partial_bid_configuration},
            )
        )
    if partial_bid_configuration and isinstance(partial_bid_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("partialBidConfiguration"),
                value={"space": partial_bid_configuration[0], "externalId": partial_bid_configuration[1]},
            )
        )
    if (
        partial_bid_configuration
        and isinstance(partial_bid_configuration, list)
        and isinstance(partial_bid_configuration[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("partialBidConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in partial_bid_configuration],
            )
        )
    if (
        partial_bid_configuration
        and isinstance(partial_bid_configuration, list)
        and isinstance(partial_bid_configuration[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("partialBidConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in partial_bid_configuration],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import itertools
import json
import logging

import pandas as pd
import yaml
from cognite.client.data_classes import Sequence

from cognite.powerops.client._generated.cogshop1 import data_classes as cogshop_v1
from cognite.powerops.resync import config
from cognite.powerops.resync.models._shared_v1_v2.cogshop_model import (
    _create_transformation,
    _create_transformation_v2,
    _to_shop_files,
    _to_shop_model_file,
)
from cognite.powerops.resync.models.base import CDFSequence
from cognite.powerops.resync.models.v1.cogshop import CogShop1Asset
from cognite.powerops.resync.models.v1.market import DayAheadProcess, RKOMProcess
from cognite.powerops.resync.models.v1.production import Watercourse

logger = logging.getLogger(__name__)


def to_cogshop_asset_model(
    configuration: config.CogShopConfig,
    watercourses: list[Watercourse],
    shop_version: str,
    dayahead_processes: list[DayAheadProcess],
    rkom_processes: list[RKOMProcess],
) -> CogShop1Asset:
    model = CogShop1Asset()

    model.shop_files.extend(_to_shop_files(configuration.watercourses_shop))

    # TODO Fix the assumption that timeseries mappings and watercourses are in the same order
    for watercourse, mapping, mapping2 in zip(
        watercourses, configuration.time_series_mappings, configuration.time_series_mappings_v2
    ):
        model_file = _to_shop_model_file(
            watercourse.name,
            watercourse.model_file,
            watercourse.processed_model_file,
            watercourse.write_back_model_file,
        )
        model.shop_files.append(model_file)

        ##### Output definition #####
        external_id = f"SHOP_{watercourse.name.replace(' ', '_')}_output_definition"

        sequence = Sequence(
            name=external_id.replace("_", " "),
            description="Defining which SHOP results to output to CDF (as time series)",
            external_id=external_id,
            columns=[
                {"valueType": "STRING", "externalId": "shop_object_type"},
                {"valueType": "STRING", "externalId": "shop_attribute_name"},
                {"valueType": "STRING", "externalId": "cdf_attribute_name"},
                {"valueType": "STRING", "externalId": "unit"},
                {"valueType": "STRING", "externalId": "is_step"},
            ],
            metadata={"shop:watercourse": watercourse.name, "shop:type": "output_definition"},
        )
        # Only default mapping is used
        default_mapping_df = pd.DataFrame(
            [
                ("market", "sale_price", "price", "EUR/MWh", "True"),
                ("market", "sale", "sales", "MWh", "True"),
                ("plant", "production", "production", "MW", "True"),
                ("plant", "consumption", "consumption", "MW", "True"),
                ("reservoir", "water_value_global_result", "water_value", "EUR/Mm3", "True"),
                ("reservoir", "energy_conversion_factor", "energy_conversion_factor", "MWh/Mm3", "True"),
            ],
            columns=[c.external_id for c in sequence.columns],
        )

        output_definition = CDFSequence(sequence=sequence, content=default_mapping_df)

        model.output_definitions.append(output_definition)

        ##### Base Mapping #####
        external_id = f"SHOP_{watercourse.name}_base_mapping"
        sequence = Sequence(
            name=external_id.replace("_", " "),
            external_id=external_id,
            description="Mapping between SHOP paths and CDF TimeSeries",
            columns=mapping.column_definitions,
            metadata={"shop:watercourse": watercourse.name, "shop:type": "base_mapping"},
        )
        output_definition = CDFSequence(sequence=sequence, content=mapping.to_dataframe())
        model.base_mappings.append(output_definition)

        ### Model Template ###
        model_template = cogshop_v1.ModelTemplateApply(
            external_id=f"ModelTemplate_{watercourse.name}",
            version="1",
            shop_version=shop_version,
            watercourse=watercourse.name,
            source="resync",
            model=cogshop_v1.FileRefApply(
                external_id=f"ModelTemplate_{watercourse.name}__FileRef_model",
                type="case",
                file_external_id=model_file.external_id,
            ),
            base_mappings=[
                cogshop_v1.MappingApply(
                    external_id=f"BM_{watercourse.name}_{row.shop_model_path}",
                    path=row.shop_model_path,
                    timeseries_external_id=row.time_series_external_id,
                    transformations=[
                        _create_transformation(order, transformation)
                        for order, transformation in enumerate(row.transformations or [])
                    ]
                    + [
                        _create_transformation_v2(order, transformation)
                        for order, transformation in enumerate(row2.transformations or [])
                    ],
                    retrieve=row.retrieve.name if row.retrieve else None,
                    aggregation=row.aggregation.name if row.aggregation else None,
                )
                for row, row2 in zip(mapping, mapping2)
            ],
        )
        model.model_templates[model_template.external_id] = model_template

    model.mappings.update(
        {
            mapping.external_id: mapping
            for template in model.model_templates.values()
            for mapping in template.base_mappings
        }
    )

    model.transformations.update(
        {
            t.external_id: t
            for template in model.model_templates.values()
            for mapping in template.base_mappings
            for t in mapping.transformations
        }
    )

    command_file_by_watercourse = {
        f.meta.metadata["shop:watercourse"]: f
        for f in model.shop_files
        if f.meta.metadata.get("shop:type") == "commands"
    }

    for process in itertools.chain(dayahead_processes, rkom_processes):
        for incremental_mapping in process.incremental_mapping:
            incremental_mapping: CDFSequence
            watercourse = incremental_mapping.sequence.metadata["shop:watercourse"]
            scenario_name = incremental_mapping.sequence.metadata["bid:scenario_name"]
            external_id = f"Scenario_{watercourse}_{scenario_name}"
            command_file = command_file_by_watercourse.get(watercourse)

            if command_file is None:
                raise ValueError(f"Could not find commands file for watercourse {watercourse}")
            scenario = cogshop_v1.ScenarioApply(
                external_id=external_id,
                name=f"Scenario {watercourse} {scenario_name}",
                model_template=model.model_templates[f"ModelTemplate_{watercourse}"].external_id,
                mappings_override=[
                    cogshop_v1.MappingApply(
                        external_id=f"Mapping_{external_id}_{i}",
                        path=mapping["shop_model_path"],
                        timeseries_external_id=mapping.get("time_series_external_id"),
                        retrieve=mapping.get("retrieve"),
                        aggregation=mapping.get("aggregation"),
                        transformations=[
                            _create_transformation(j, transformation_data)
                            for j, transformation_data in enumerate(json.loads(mapping.get("transformations", "")))
                        ]
                        + [
                            _create_transformation_v2(j, config.Transformation(**transformation_data))
                            for j, transformation_data in enumerate(json.loads(mapping.get("transformations", "")))
                        ],
                    )
                    for i, mapping in enumerate(
                        incremental_mapping.content.replace(float("nan"), None).to_dict(orient="records")
                    )
                ],
                commands=cogshop_v1.CommandsConfigApply(
                    external_id=f"Commands_{watercourse}",
                    commands=yaml.safe_load(command_file.content.decode()).get("commands", []),
                ),
                source="resync",
            )
            model.scenarios[scenario.external_id] = scenario

    model.mappings.update(
        {
            mapping.external_id: mapping
            for scenario in model.scenarios.values()
            for mapping in scenario.mappings_override
        }
    )
    model.transformations.update(
        {
            t.external_id: t
            for scenario in model.scenarios.values()
            for mapping in scenario.mappings_override
            for t in mapping.transformations
        }
    )
    model.commands_configs.update(
        {scenario.commands.external_id: scenario.commands for scenario in model.scenarios.values()}
    )

    return model



from __future__ import annotations

import logging
from pathlib import Path
from typing import Annotated, Optional

import typer
from rich.logging import Console, RichHandler

from cognite import powerops
from cognite.powerops.utils.cdf.extraction_pipelines import ExtractionPipelineCreate, RunStatus

for third_party in ["cognite-sdk", "requests", "urllib3", "msal", "requests_oauthlib"]:
    third_party_logger = logging.getLogger(third_party)
    third_party_logger.setLevel(logging.WARNING)
    third_party_logger.propagate = False

MODELS_BY_NAME_SORTED = sorted(powerops.resync.MODELS_BY_NAME)
MODELS_BY_NAME_STR = ", ".join(MODELS_BY_NAME_SORTED)

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.INFO, format=FORMAT, datefmt="[%X]", handlers=[RichHandler(console=Console(stderr=True))]
)

log = logging.getLogger("rich")

app = typer.Typer(pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)


def _version_callback(value: bool):
    if value:
        typer.echo(powerops.__version__)
        raise typer.Exit()


@app.callback()
def common(ctx: typer.Context, version: bool = typer.Option(None, "--version", callback=_version_callback)): ...


@app.command("init", help="Setup necessary data models in CDF for ReSync to run")
def init(
    dev: bool = typer.Option(
        False,
        "--dev",
        "-d",
        help="Whether the deployment is for development environment. If true,"
        "the models views and data models will be deleted and recreated.",
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Whether to run the command as a dry run, meaning no resources will be created."
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Whether to print verbose output"),
):
    client = powerops.PowerOpsClient.from_settings()

    powerops.resync.init(client, is_dev=dev, dry_run=dry_run, verbose=verbose)


@app.command("validate", help="Validate the configuration files and timeseries")
def validate(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Include valid results in output (specify --format), and enable INFO logs"
    ),
    format: str = typer.Option(default=None, help="The format of the output. Available formats: markdown, json"),
    error_code: bool = typer.Option(False, "--error-code", "-e", help="Exit with error code if validation fails"),
):
    echo = _setup_echo(verbose, typer.echo)
    log.info(f"Running validate on configuration files located in {path}")
    validation_results = powerops.resync.validate(path, market)

    if format == "markdown":
        echo(validation_results.as_markdown(include_valid=verbose))
    elif format == "json":
        echo(validation_results.as_json(include_valid=verbose))

    if error_code and not all(result.valid for result in validation_results.results):
        raise typer.Exit(code=1)


@app.command(
    "plan",
    help="Preview the changes from the configuration files that `powerops apply` would make to the data model in CDF",
)
def plan(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
    models: list[str] = typer.Option(  # noqa: B008
        default=MODELS_BY_NAME_SORTED,
        help=f"The models to run the plan. Available models: {MODELS_BY_NAME_STR}",
    ),
    dump_folder: Optional[Path] = typer.Option(  # noqa: B008
        default=None, help="If present, the local and cdf changes will be dumped to this directory."
    ),
    format: str = typer.Option(default=None, help="The format of the output. Available formats: markdown"),
    as_extraction_pipeline_run: bool = typer.Option(
        default=False,
        help="If true, the command will be registered as an extraction pipeline run. With the configuration"
        "fetched from the settings.toml [powerops] section.",
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Whether to print verbose output"),
):
    if dump_folder and not dump_folder.is_dir():
        raise typer.BadParameter(f"{dump_folder} is not a directory")

    echo = _setup_echo(verbose, typer.echo)
    log.info(f"Running plan on configuration files located in {path}")
    power = powerops.PowerOpsClient.from_settings()

    changes = powerops.resync.plan(path, market, model_names=models, dump_folder=dump_folder, client=power)

    if format == "markdown":
        echo(changes.as_github_markdown())

    if as_extraction_pipeline_run is True:
        client = power.cdf

        pipeline = ExtractionPipelineCreate(
            external_id="resync/plan",
            data_set_external_id=power.datasets.monitor_dataset,
            dump_truncated_to_file=True,
            message_keys_skip=["error"],
            truncate_keys_first=["error"],
            log_file_prefix="powerops_function_loss",
            description="The resync/plan function checks that the configuration files are matching "
            "the expected resources in CDF. If there are any differences, the run will report as failed",
        ).get_or_create(client)

        with pipeline.create_pipeline_run(client) as run:
            if changes.has_changes("CDF", exclude={"timeseries", "parent_assets", "labels"}):
                run.update_data(
                    RunStatus.FAILURE,
                    summary=changes.as_markdown_summary(no_headers=True),
                    error=changes.as_markdown_detailed(),
                )
            else:
                run.update_data(RunStatus.SUCCESS)
        echo(f"Extraction pipeline run executed with status: {run.status}")


@app.command("apply", help="Apply the changes from the configuration files to the data model in CDF")
def apply(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
    models: list[str] = typer.Option(  # noqa: B008
        default=MODELS_BY_NAME_SORTED,
        help=f"The models to run apply. Available models: {MODELS_BY_NAME_STR}",
    ),
    auto_yes: bool = typer.Option(False, "--yes", "-y", help="Auto confirm all prompts"),
    format: str = typer.Option(default=None, help="The format of the output. Available formats: markdown"),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Whether to print verbose output"),
):
    echo = _setup_echo(verbose, typer.echo)
    client = powerops.PowerOpsClient.from_settings()

    logging.info(f"Running apply on configuration files located in {path}")
    changed = powerops.resync.apply(path, market, client, model_names=models, auto_yes=auto_yes)
    if format == "markdown":
        echo(changed.as_github_markdown())


@app.command("destroy", help="Destroy all the data models created by resync and remove all the data.")
def destroy(
    models: list[str] = typer.Option(  # noqa: B008
        default=MODELS_BY_NAME_SORTED,
        help=f"The models to destroy. Available models: {MODELS_BY_NAME_STR}",
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Whether to run the command as a dry run, meaning no resources will be destroyed."
    ),
    auto_yes: bool = typer.Option(False, "--yes", "-y", help="Auto confirm all prompts"),
    format: str = typer.Option(default=None, help="The format of the output. Available formats: markdown"),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Whether to print verbose output"),
):
    echo = _setup_echo(verbose, typer.echo)
    client = powerops.PowerOpsClient.from_settings()

    destroyed = powerops.resync.destroy(client, model_names=models, auto_yes=auto_yes, dry_run=dry_run)
    if format == "markdown":
        echo(destroyed.as_github_markdown())


@app.command("plan_v1", help="Plan the changes from the configuration files to the data model in CDF")
def plan_v1(path: Annotated[Path, typer.Argument(help="Path to configuration files")]):
    logging.info(f"Running plan on configuration files located in {path}")
    powerops.resync.plan_v1(path)


@app.command("apply_v1", help="Apply the changes from the configuration files to the data model in CDF")
def apply_v1(path: Annotated[Path, typer.Argument(help="Path to configuration files")]):
    logging.info(f"Running apply on configuration files located in {path}")
    powerops.resync.apply_v1(path)


def main():
    app()


def _setup_echo(verbose: bool, echo_func: Optional[powerops.resync.Echo] = None) -> powerops.resync.Echo:
    if not verbose:
        logging.disable(logging.INFO)  # disable logs for INFO and below

    return echo_func or print


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import ClassVar, Union

from cognite.client.data_classes import TimeSeries
from pydantic import Field, field_validator, model_serializer, model_validator

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import NonAssetType
from cognite.powerops.utils.serialization import try_load_dict, try_load_list

from .base import Bid, Process, ShopTransformation


class BenchmarkBid(Bid):
    market_config_external_id: str


class ProductionPlanTimeSeries(NonAssetType):
    name: str
    series: list[TimeSeries]

    @field_validator("series", mode="after")
    @classmethod
    def ordering(cls, value: list[TimeSeries]) -> list[TimeSeries]:
        return sorted(value, key=lambda x: x.external_id)

    @model_validator(mode="before")
    @classmethod
    def parse_dict(cls, value) -> dict:
        if isinstance(value, dict):
            name, series = next(iter(value.items()))
            return {"name": name, "series": [TimeSeries(external_id=s) for s in series]}
        return value

    @model_serializer()
    def ser_dict(self) -> dict:
        return {self.name: [s.external_id for s in self.series]}

    def standardize(self) -> None:
        self.series = self.ordering(self.series)


class BenchmarkProcess(Process):
    parent_external_id: ClassVar[str] = "benchmarking_configurations"
    parent_description: ClassVar[str] = "Configurations used in benchmarking processes"
    description: str = "Configuration for benchmarking of day-ahead bidding"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.DAYAHEAD_BIDDING_BENCHMARKING_CONFIG
    shop: ShopTransformation
    bid: BenchmarkBid
    production_plan_time_series: list[ProductionPlanTimeSeries] = Field(default_factory=list)
    benchmarking_metrics: dict[str, str] = Field(default_factory=dict)
    run_events: list[str] = Field(default_factory=list)
    bid_process_configuration_assets: list[Process] = Field(default_factory=list)

    @field_validator("production_plan_time_series", mode="after")
    @classmethod
    def ordering(cls, value: list[ProductionPlanTimeSeries]) -> list[ProductionPlanTimeSeries]:
        return sorted(value, key=lambda x: x.name)

    @field_validator("run_events", mode="after")
    @classmethod
    def ordering_events(cls, value: list[str]) -> list[str]:
        return sorted(value)

    @field_validator("bid_process_configuration_assets", mode="after")
    @classmethod
    def ordering_processes(cls, value: list[Process]) -> list[Process]:
        return sorted(value, key=lambda x: x.name)

    @field_validator("benchmarking_metrics", mode="before")
    @classmethod
    def parse_str(cls, value) -> dict:
        return try_load_dict(value)

    @field_validator("production_plan_time_series", mode="before")
    @classmethod
    def parse_str_to_list(cls, value) -> list:
        if isinstance(loaded := try_load_list(value), dict):
            return [loaded]
        return loaded

    def standardize(self) -> None:
        self.production_plan_time_series = self.ordering(self.production_plan_time_series)
        self.run_events = self.ordering_events(self.run_events)
        self.bid_process_configuration_assets = self.ordering_processes(self.bid_process_configuration_assets)



from .echo import Echo
from .main import DATAMODEL_ID_TO_RESYNC_NAME, MODELS_BY_NAME, apply, destroy, init, plan, validate

__all__ = ["apply", "plan", "init", "destroy", "validate", "MODELS_BY_NAME", "Echo", "DATAMODEL_ID_TO_RESYNC_NAME"]



from __future__ import annotations

from typing import Optional, TypeVar

from pydantic import BaseModel, Field

T = TypeVar("T")


class Connection(BaseModel):
    connection_type: str
    from_: str = Field(alias="from")
    from_type: Optional[str] = None
    to: str
    to_type: Optional[str] = None

    def to_from_any(self, from_: dict[str, T], to_name: str, to_type: str) -> T | None:
        if self.to != to_name or self.to_type != to_type:
            return None
        f = from_.get(self.from_)
        if f:
            return f
        return None

    def from_to_any(self, from_name: str, from_type: str, to: dict[str, T]) -> T | None:
        if self.from_ != from_name or self.from_type != from_type:
            return None
        f = to.get(self.to)

        if f:
            return f
        return None



from __future__ import annotations

from abc import ABC
from typing import ClassVar, Union

from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import AssetType, NonAssetType


class Market(AssetType):
    parent_external_id: ClassVar[str] = "market_configurations"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.MARKET
    parent_description: ClassVar[str] = "Configurations used for different markets"
    timezone: str


class DateTransformation(NonAssetType):
    transformation: str
    args: list[Union[int, float, str]] = Field(default_factory=list)
    kwargs: dict[str, Union[int, float, str]] = Field(default_factory=dict)


class ShopTransformation(NonAssetType):
    starttime: str
    endtime: str


class Bid(NonAssetType):
    date: str


class Process(AssetType, ABC):
    ...
    # TODO: extract common fields
    # parent_external_id: ClassVar[str]  # noqa: ERA001
    # parent_description: ClassVar[str]  # noqa: ERA001
    # label: ClassVar[Union[AssetLabel, str]]  # noqa: ERA001
    # shop: ShopTransformation  # noqa: ERA001
    # bid: Bid  # noqa: ERA001
    # incremental_mapping: list[CDFSequence] = Field(default_factory=list)  # noqa: ERA001
    # (the noqa is for ruff to avoid removing these comments -_- )



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.day_ahead_bid.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.day_ahead_bid.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    PriceArea,
    PriceAreaWrite,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaWriteList,
    PriceAreaTextFields,
)
from cognite.powerops.client._generated.day_ahead_bid.data_classes._price_area import (
    _PRICEAREA_PROPERTIES_BY_FIELD,
    _create_price_area_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .price_area_main_scenario import PriceAreaMainScenarioAPI
from .price_area_price_scenarios import PriceAreaPriceScenariosAPI
from .price_area_query import PriceAreaQueryAPI


class PriceAreaAPI(NodeAPI[PriceArea, PriceAreaWrite, PriceAreaList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[PriceArea]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=PriceArea,
            class_list=PriceAreaList,
            class_write_list=PriceAreaWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.main_scenario = PriceAreaMainScenarioAPI(client, view_id)
        self.price_scenarios = PriceAreaPriceScenariosAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> PriceAreaQueryAPI[PriceAreaList]:
        """Query starting at price areas.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            default_method: The default method to filter on.
            timezone: The timezone to filter on.
            timezone_prefix: The prefix of the timezone to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for price areas.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            default_method,
            timezone,
            timezone_prefix,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(PriceAreaList)
        return PriceAreaQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        price_area: PriceAreaWrite | Sequence[PriceAreaWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) price areas.

        Args:
            price_area: Price area or sequence of price areas to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new price_area:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> from cognite.powerops.client._generated.day_ahead_bid.data_classes import PriceAreaWrite
                >>> client = DayAheadBidAPI()
                >>> price_area = PriceAreaWrite(external_id="my_price_area", ...)
                >>> result = client.price_area.apply(price_area)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.price_area.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(price_area, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more price area.

        Args:
            external_id: External id of the price area to delete.
            space: The space where all the price area are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete price_area by id:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> client.price_area.delete("my_price_area")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.price_area.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> PriceArea | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> PriceAreaList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PriceArea | PriceAreaList | None:
        """Retrieve one or more price areas by id(s).

        Args:
            external_id: External id or list of external ids of the price areas.
            space: The space where all the price areas are located.

        Returns:
            The requested price areas.

        Examples:

            Retrieve price_area by id:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> price_area = client.price_area.retrieve("my_price_area")

        """
        return self._retrieve(external_id, space)

    def search(
        self,
        query: str,
        properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PriceAreaList:
        """Search price areas

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            default_method: The default method to filter on.
            timezone: The timezone to filter on.
            timezone_prefix: The prefix of the timezone to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results price areas matching the query.

        Examples:

           Search for 'my_price_area' in all text properties:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> price_areas = client.price_area.search('my_price_area')

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            default_method,
            timezone,
            timezone_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _PRICEAREA_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: PriceAreaFields | Sequence[PriceAreaFields] = None,
        query: str | None = None,
        search_properties: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        group_by: PriceAreaFields | Sequence[PriceAreaFields] | None = None,
        query: str | None = None,
        search_property: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across price areas

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            default_method: The default method to filter on.
            timezone: The timezone to filter on.
            timezone_prefix: The prefix of the timezone to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count price areas in space `my_space`:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> result = client.price_area.aggregate("count", space="my_space")

        """

        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            default_method,
            timezone,
            timezone_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PRICEAREA_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: PriceAreaFields,
        interval: float,
        query: str | None = None,
        search_property: PriceAreaTextFields | Sequence[PriceAreaTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for price areas

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            default_method: The default method to filter on.
            timezone: The timezone to filter on.
            timezone_prefix: The prefix of the timezone to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            default_method,
            timezone,
            timezone_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PRICEAREA_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        default_method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PriceAreaList:
        """List/filter price areas

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            default_method: The default method to filter on.
            timezone: The timezone to filter on.
            timezone_prefix: The prefix of the timezone to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of price areas to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            List of requested price areas

        Examples:

            List price areas and limit to 5:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> price_areas = client.price_area.list(limit=5)

        """
        filter_ = _create_price_area_filter(
            self._view_id,
            name,
            name_prefix,
            default_method,
            timezone,
            timezone_prefix,
            external_id_prefix,
            space,
            filter,
        )
        return self._list(limit=limit, filter=filter_)



"""
This package contains parent classes that all other models inherit from. These parent classes provide convenience
methods for working with the models, and also provide a standardized interface for the models.
"""

from .asset_model import AssetModel
from .asset_type import AssetType, NonAssetType, T_Asset_Type
from .cdf_resources import CDFFile, CDFResource, CDFSequence, SpaceId
from .data_model import DataModel
from .graph_ql import PowerOpsGraphQLModel
from .model import Model, T_Model
from .resource_type import ResourceType

__all__ = [
    "AssetModel",
    "AssetType",
    "T_Asset_Type",
    "NonAssetType",
    "DataModel",
    "Model",
    "T_Model",
    "CDFSequence",
    "CDFFile",
    "CDFResource",
    "PowerOpsGraphQLModel",
    "ResourceType",
    "SpaceId",
]



from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal, Protocol, TypeVar, Union, cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, FileMetadata, FileMetadataList, LabelDefinition, Relationship
from cognite.client.data_classes import Sequence as CogniteSequence
from cognite.client.data_classes.data_modeling import ContainerId, DataModelId, EdgeApply, NodeApply, ViewId
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.powerops.resync.models.base import CDFFile, CDFSequence
from cognite.powerops.utils.navigation import chunks

T_CogniteResource = TypeVar(
    "T_CogniteResource",
    bound=Union[
        Asset,
        CogniteSequence,
        FileMetadata,
        Relationship,
        NodeApply,
        EdgeApply,
        ContainerId,
        ViewId,
        DataModelId,
        LabelDefinition,
    ],
)


class CogniteAPI(Protocol[T_CogniteResource]):  # type: ignore[misc]
    def create(self, items: T_CogniteResource | Sequence[T_CogniteResource]) -> Any: ...

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any: ...

    def upsert(
        self, item: T_CogniteResource | Sequence[T_CogniteResource], mode: Literal["patch", "replace"] = "patch"
    ) -> Any: ...


class FileAdapter(CogniteAPI[FileMetadata]):
    def __init__(self, client: CogniteClient, files_by_id: dict[str, CDFFile] | None = None):
        self.client = client
        self.files_by_id = files_by_id

    def create(self, items: FileMetadata | Sequence[FileMetadata]) -> Any:
        if self.files_by_id is None:
            raise ValueError("Missing file content need to be provided")
        items = [items] if isinstance(items, FileMetadata) else items
        for item in items:
            if item.external_id is None or item.external_id not in self.files_by_id:
                raise ValueError("Cannot create new file {item.external_id} missing file content")
            content = self.files_by_id[item.external_id].content
            if content is None:
                raise ValueError(f"Cannot create new file {item.external_id} missing file content")
            self.client.files.upload_bytes(content, **item.dump(camel_case=False), overwrite=True)

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any:
        self.client.files.delete(external_id=external_id)

    def upsert(self, item: FileMetadata | Sequence[FileMetadata], mode: Literal["patch", "replace"] = "patch") -> Any:
        items = [item] if isinstance(item, FileMetadata) else item
        updated = self.client.files.update(items)
        update_by_id = {u.external_id: u for u in ([updated] if isinstance(updated, FileMetadata) else updated)}

        # There are not all fields that can be updates for files, for example, name.
        # Therefore, we need to delete and recreate files that have changed for example name.
        to_delete_and_recreate = FileMetadataList([])
        for item in items:
            if item.external_id not in update_by_id:
                raise ValueError(f"Could not update file {item.external_id}")
            if item != update_by_id[item.external_id]:
                to_delete_and_recreate.append(item)
        if to_delete_and_recreate:
            self.delete(to_delete_and_recreate.as_external_ids())
            self.create(to_delete_and_recreate)


class SequenceAdapter(CogniteAPI[CogniteSequence]):
    def __init__(self, client: CogniteClient, sequence_by_id: dict[str, CDFSequence] | None = None):
        self.client = client
        self.sequence_by_id = sequence_by_id

    def create(self, items: CogniteSequence | Sequence[CogniteSequence]) -> Any:
        if self.sequence_by_id is None:
            raise ValueError("Missing sequence content need to be provided")

        items = [items] if isinstance(items, CogniteSequence) else items
        if missing := {i.external_id for i in items} - set(self.sequence_by_id):
            raise ValueError(f"Missing sequence content for {missing}")
        self.client.sequences.upsert(items, mode="replace")
        for item in items:
            if item.external_id is None:
                raise ValueError("Missing external id for sequence")
            cdf_sequence_df = self.sequence_by_id[item.external_id].content
            if cdf_sequence_df is None:
                raise ValueError(f"Missing sequence content for {item.external_id}")
            self.client.sequences.data.insert_dataframe(cdf_sequence_df, external_id=item.external_id)

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any:
        self.client.sequences.delete(external_id=external_id)

    def upsert(
        self, item: CogniteSequence | Sequence[CogniteSequence], mode: Literal["patch", "replace"] = "patch"
    ) -> Any:
        self.client.sequences.update(item)


T_Instance = TypeVar("T_Instance", bound=Union[NodeApply, EdgeApply])


class InstanceAdapter(CogniteAPI[T_Instance]):
    instance_limit = 1000

    def __init__(self, client: CogniteClient, instance_type: Literal["node", "edge"]):
        self.instance_type = instance_type
        self.client = client

    def create(self, items: T_Instance | Sequence[T_Instance]) -> Any:
        item_sequence: Sequence[T_Instance]
        if not isinstance(items, Sequence):
            item_sequence = [items]
        else:
            item_sequence = items

        failed = []
        for chunk in chunks(item_sequence, 500):  # type: ignore[type-var]
            if self.instance_type == "node":
                try:
                    self.client.data_modeling.instances.apply(nodes=chunk, auto_create_direct_relations=True)  # type: ignore[arg-type]
                except CogniteAPIError:
                    print(f"Failed on {chunk[0].dump()}")
                    failed.append(chunk[0].dump())
            else:
                try:
                    self.client.data_modeling.instances.apply(
                        edges=chunk,  # type: ignore[arg-type]
                        auto_create_start_nodes=True,
                        auto_create_end_nodes=True,
                    )
                except CogniteAPIError:
                    print(f"Failed on {chunk[0].dump()}")
                    failed.append(chunk[0].dump())
        if failed:
            raise ValueError(f"Failed on {len(failed)}/{len(item_sequence)}")

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any:
        if self.instance_type == "node":
            self.client.data_modeling.instances.delete(nodes=external_id)  # type: ignore[arg-type]
        else:
            self.client.data_modeling.instances.delete(edges=external_id)  # type: ignore[arg-type]

    def upsert(self, item: T_Instance | Sequence[T_Instance], mode: Literal["patch", "replace"] = "patch") -> Any:
        self.create(item)


class DataModelingAdapter(CogniteAPI[T_CogniteResource]):
    def __init__(self, api: Any) -> None:
        self.api = api

    def create(self, items: T_CogniteResource | Sequence[T_CogniteResource]) -> Any:
        self.api.apply(items)

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any:
        self.api.delete(external_id)

    def upsert(
        self, item: T_CogniteResource | Sequence[T_CogniteResource], mode: Literal["patch", "replace"] = "patch"
    ) -> Any:
        self.api.apply(item)


class LabelsAdapter(CogniteAPI[LabelDefinition]):
    def __init__(self, client: CogniteClient):
        self.client = client

    def create(self, items: LabelDefinition | Sequence[LabelDefinition]) -> Any:
        self.client.labels.create(items)

    def delete(self, external_id: str | SequenceNotStr[str]) -> Any:
        # Labels are not deleted.
        ...

    def upsert(
        self, item: LabelDefinition | Sequence[LabelDefinition], mode: Literal["patch", "replace"] = "patch"
    ) -> Any:
        # Labels are not changes
        ...


def get_cognite_api(
    client: CogniteClient,
    name: str,
    new_sequences_by_id: dict[str, CDFSequence] | None = None,
    new_files_by_id: dict[str, CDFFile] | None = None,
) -> CogniteAPI:
    if name == "assets" or name == "parent_assets":
        return cast(CogniteAPI[Asset], client.assets)
    elif name == "labels":
        return LabelsAdapter(client)
    elif name == "time_series":
        raise NotImplementedError("Resync does not create timeseries")
    elif name == "sequences":
        return SequenceAdapter(client, new_sequences_by_id)
    elif name == "files":
        return FileAdapter(client, new_files_by_id)
    elif name == "relationships":
        return client.relationships
    elif name == "nodes":
        return InstanceAdapter[NodeApply](client, "node")
    elif name == "edges":
        return InstanceAdapter[EdgeApply](client, "edge")
    elif name == "data_models":
        return DataModelingAdapter[DataModelId](client.data_modeling.data_models)
    elif name == "views":
        return DataModelingAdapter[ViewId](client.data_modeling.views)
    elif name == "containers":
        return DataModelingAdapter[ContainerId](client.data_modeling.containers)
    raise ValueError(f"Unknown resource type {name}")



from __future__ import annotations

from typing import ClassVar, Optional, Union

from pydantic import Field, field_validator

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import AssetType, CDFSequence, NonAssetType
from cognite.powerops.utils.serialization import try_load_list

from .base import Bid, Market, Process, ShopTransformation


class RKOMBid(Bid):
    auction: str
    block: str
    method: str
    minimum_price: str
    price_premium: str
    product: str
    watercourse: str
    price_scenarios: str
    reserve_scenarios: str


class RKOMPlants(NonAssetType):
    plants: str


class RKOMProcess(Process):
    parent_external_id: ClassVar[str] = "rkom_bid_process_configurations"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.RKOM_BID_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations used in RKOM bid generation processes"
    shop: ShopTransformation
    bid: RKOMBid
    process_events: list[str] = Field(default_factory=list)
    timezone: str
    rkom: RKOMPlants
    incremental_mapping: list[CDFSequence] = Field(default_factory=list)

    @field_validator("incremental_mapping", mode="after")
    @classmethod
    def ordering(cls, value: list[CDFSequence]) -> list[CDFSequence]:
        return sorted(value, key=lambda x: x.external_id)

    @field_validator("process_events", mode="before")
    @classmethod
    def ordering_events(cls, value: list[str]) -> list[str]:
        return sorted(value)

    @field_validator("process_events", mode="before")
    @classmethod
    def parse_str(cls, value) -> list:
        return try_load_list(value)

    def standardize(self) -> None:
        self.incremental_mapping = self.ordering(self.incremental_mapping)
        self.process_events = self.ordering_events(self.process_events)


class RKOMCombinationBid(NonAssetType):
    auction: str
    combination_name: str
    rkom_bid_configs: list[str]

    @field_validator("rkom_bid_configs", mode="before")
    @classmethod
    def parse_str(cls, value) -> list:
        return try_load_list(value)

    @field_validator("rkom_bid_configs", mode="after")
    @classmethod
    def ordering(cls, value: list[str]) -> list[str]:
        return sorted(value)

    def standardize(self) -> None:
        self.rkom_bid_configs = self.ordering(self.rkom_bid_configs)


class RKOMBidCombination(AssetType):
    parent_external_id: ClassVar[str] = "rkom_bid_combination_configurations"
    label: ClassVar[str] = AssetLabel.RKOM_BID_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations for which bids should be combined into a total RKOM bid form"
    bid: RKOMCombinationBid

    def standardize(self) -> None:
        self.bid.standardize()


class RKOMMarket(Market):
    # Temporary optional to fix that this is missing in CDF.
    start_of_week: Optional[int] = None



from ._main import CogShopConfig, MarketConfig, ProductionConfig, ReSyncConfig
from ._settings import Settings
from ._shared import (
    ATTRIBUTE_DEFAULT_AGGREGATION,
    AggregationMethod,
    Auction,
    RetrievalType,
    TimeSeriesMapping,
    TimeSeriesMappingEntry,
    Transformation,
    TransformationType,
)
from .cogshop import ShopFileConfig
from .market import (
    MARKET_BY_PRICE_AREA,
    MARKET_CONFIG_NORDPOOL_DAYAHEAD,
    BenchmarkingConfig,
    BidMatrixGeneratorConfig,
    BidProcessConfig,
    Block,
    Market,
    PriceScenario,
    PriceScenarioID,
    PriceScenarioV2,
    Product,
    ReserveScenario,
    ReserveScenarios,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    RkomMarketConfig,
)
from .production import (
    Connection,
    Generator,
    GeneratorTimeSeriesMapping,
    Plant,
    PlantTimeSeriesMapping,
    Watercourse,
    WatercourseConfig,
)

__all__ = [
    "Settings",
    "MarketConfig",
    "ProductionConfig",
    "ReSyncConfig",
    "CogShopConfig",
    "PlantTimeSeriesMapping",
    "Plant",
    "WatercourseConfig",
    "Watercourse",
    "Connection",
    "ShopFileConfig",
    "GeneratorTimeSeriesMapping",
    "Generator",
    "MARKET_BY_PRICE_AREA",
    "PriceScenario",
    "PriceScenarioV2",
    "PriceScenarioID",
    "BenchmarkingConfig",
    "ReserveScenario",
    "ReserveScenarios",
    "RKOMBidCombinationConfig",
    "RKOMBidProcessConfig",
    "RkomMarketConfig",
    "Product",
    "Block",
    "BidMatrixGeneratorConfig",
    "BidProcessConfig",
    "Market",
    "MARKET_CONFIG_NORDPOOL_DAYAHEAD",
    "TimeSeriesMapping",
    "TimeSeriesMappingEntry",
    "Transformation",
    "Auction",
    "RetrievalType",
    "AggregationMethod",
    "TransformationType",
    "ATTRIBUTE_DEFAULT_AGGREGATION",
]



from __future__ import annotations

import json
import logging
import traceback
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Callable, ClassVar, Optional

from cognite.client import CogniteClient
from cognite.client._constants import MAX_VALID_INTERNAL_ID
from cognite.client.data_classes import ExtractionPipeline, ExtractionPipelineRun
from cognite.client.exceptions import CogniteAPIError

from cognite.powerops.utils.retry import retry

logger = logging.getLogger(__name__)

MSG_CHAR_LIMIT = 1000


class RunStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    SEEN = "seen"


@dataclass
class _Config:
    dump_truncated_to_file: bool = True
    message_keys_skip: list[str] = field(default_factory=list)
    truncate_keys: list[str] = field(default_factory=list)
    reverse_truncate_keys: list[str] = field(default_factory=list)
    log_file_prefix: str | None = None


class PipelineRun:
    """
    A context manager to create an extraction pipeline run when exiting the context.
    Run status defaults to SEEN and FAILURE if an exception is raised. To set a custom
    status and message, use the `update_data` method.

    """

    init_status: ClassVar[RunStatus] = RunStatus.SEEN
    log_file_external_id: ClassVar[str] = "log_file_external_id"
    log_file_id: ClassVar[str] = "log_file_id"
    exception: ClassVar[str] = "exception"

    reserved_keys: ClassVar[tuple[str, ...]] = (log_file_external_id, log_file_id, exception)

    def __init__(
        self,
        client: CogniteClient,
        pipeline_external_id: str,
        config: _Config,
        data_set_id: int,
        error_logger: Callable[[str], None],
        is_try_run: bool,
    ) -> None:
        self.client = client
        self.pipeline_external_id = pipeline_external_id
        self.config = config
        self.data_set_id = data_set_id
        self.error_logger = error_logger
        self.is_dry_run = is_try_run

        self.data: dict[str, Any] = {}
        self.status = self.init_status

    def __enter__(self) -> PipelineRun:
        return self

    def update_data(self, status: RunStatus | None = None, **data: dict[str, Any]) -> PipelineRun:
        if status is not None:
            self.status = status
        if isinstance(data, dict):
            for key in self.reserved_keys:
                if key in data:
                    self.error_logger(f"Key {key} is reserved and cannot be used in data. This will be skipped")
                    del data[key]
            self.data.update(data)
        else:
            raise TypeError(f"Data must be a dict, got {type(data)}")
        return self

    def __exit__(self, exc_type=None, exc_value=None, traceback_=None) -> bool:  # type: ignore[no-untyped-def]
        suppress_exception = False

        if any((exc_type, exc_value, traceback_)):
            self.status = RunStatus.FAILURE
            self.data[self.exception] = traceback.format_exc()

        message = self.get_message(self.config.dump_truncated_to_file)
        run = ExtractionPipelineRun(
            status=self.status.value, extpipe_external_id=self.pipeline_external_id, message=message
        )
        if not self.is_dry_run:
            retry(tries=5, delay=1, backoff=3, logger=logger)(self.client.extraction_pipelines.runs.create)(run)

        return suppress_exception

    def get_message(self, dump_truncated_to_file: bool) -> str:
        file_external_id = (
            f"{self.config.log_file_prefix}/{self.pipeline_external_id}/"
            f"{datetime.now(timezone.utc).isoformat().replace(':', '')}"
        )
        data, file_content = self._create_run_data_and_file_content(file_external_id, dump_truncated_to_file)
        if dump_truncated_to_file and file_content and not self.is_dry_run:
            try:
                file_id = self.client.files.upload_bytes(
                    content=file_content,
                    name=f"{file_external_id.lower()}.log".replace("/", "_"),
                    external_id=file_external_id,
                    mime_type="text/plain",
                    data_set_id=self.data_set_id,
                ).id
            except CogniteAPIError as e:
                self.error_logger(f"Failed to upload log file {file_external_id}: {e}")
                file_id = None

            data[self.log_file_id] = file_id

        return self._as_json(data, self.config.message_keys_skip)

    def _create_run_data_and_file_content(
        self, file_external_id: str, dump_truncated_to_file: bool
    ) -> tuple[dict, str]:
        file_content = []
        data = self.data.copy()
        if dump_truncated_to_file:
            data[self.log_file_external_id] = file_external_id
            # Need a dummy id as we do not know the id before the file is uploaded
            data[self.log_file_id] = MAX_VALID_INTERNAL_ID
        # Ensure we get a dict[str, str]
        for key in list(data):
            value = data[key]
            data[key] = self._as_json(value) if isinstance(value, (dict, list)) else str(value)

        truncating = len(self._as_json(data)) - MSG_CHAR_LIMIT > 0
        truncate_keys = self.config.truncate_keys + list(
            set(data) - {self.log_file_id, self.log_file_external_id} - set(self.config.truncate_keys)
        )
        while truncate_keys and (above_limit := len(self._as_json(data)) - MSG_CHAR_LIMIT) > 0:
            # In case, truncating the specified keys is not enough, we also start to truncate everything else.
            key = truncate_keys.pop(0)
            if key not in data:
                continue

            entry_length = len(data[key])
            if entry_length < 3:
                continue

            if entry_length >= above_limit + 3:
                trim_len = entry_length - (above_limit + 3)
                if key in self.config.reverse_truncate_keys:
                    data[key] = "..." + data[key][-trim_len:]
                else:
                    data[key] = data[key][:trim_len] + "..."
            else:
                data[key] = "..."

        if truncating:
            # We dump all data to the file, even the keys which are not truncated.
            for key, value in self.data.items():
                if key in {self.log_file_id, self.log_file_external_id}:
                    continue
                file_content.append(f"{'='*70}\n{key}\n{'='*70}\n{value}\n")

        return data, "\n\n".join(file_content)

    @staticmethod
    def _as_json(data: dict | list, exclude_keys: Iterable[str] | None = None) -> str:
        exclude_keys = set(exclude_keys or [])
        cleaned: dict[str, Any] | list[Any]
        # Removing space to use as little space as possible
        if isinstance(data, dict):
            cleaned = {k: v for k, v in data.items() if k not in exclude_keys}
        elif isinstance(data, list):
            cleaned = data
        else:
            raise TypeError(f"Data must be a dict or list, got {type(data)}")
        return json.dumps(cleaned, separators=(",", ":"), default=str)


class ExtractionPipelineCreate:
    """
    This is a simplified write version of an Extraction pipeline.

    This ensures that the Extraction Pipeline exists when calling the `create_pipeline_run` method.

    Args:
        external_id: The external id of the extraction pipeline.
        data_set_external_id: The external id of the dataset to use.
        description: The description of the extraction pipeline.
        dump_truncated_to_file: Whether to dump truncated data to a file. This is used when the data is too large to
            be stored in the message field of the extraction pipeline run.
        message_keys_skip: The keys that should not be part of the pipeline run message, and instead only
            be dumped to a file.
        truncate_keys_first: The keys to truncate first. This is useful when you expect the data to be too large too
            to be stored in a message field of the extraction pipeline run, and you want to select which keys
            to truncate first.
        reverse_truncate_keys: Keys with important content at the end, so trim the start of the content.
    """

    def __init__(
        self,
        external_id: str,
        data_set_external_id: str,
        description: str | None = None,
        dump_truncated_to_file: bool = True,
        message_keys_skip: Optional[list[str]] = None,
        truncate_keys_first: Optional[list[str]] = None,
        reverse_truncate_keys: Optional[list[str]] = None,
        log_file_prefix: str | None = None,
    ) -> None:
        self.external_id = external_id
        self.dataset_external_id = data_set_external_id
        self.description = description
        self._data_set_id: Optional[int] = None

        self.config = _Config(
            dump_truncated_to_file=dump_truncated_to_file,
            message_keys_skip=message_keys_skip or [],
            truncate_keys=truncate_keys_first or [],
            reverse_truncate_keys=reverse_truncate_keys or [],
            log_file_prefix=log_file_prefix,
        )

    def get_or_create(self, client: CogniteClient) -> ExtractionPipelineCreate:
        extraction_pipeline = client.extraction_pipelines.retrieve(external_id=self.external_id)
        if extraction_pipeline is None:
            self._data_set_id = self._get_data_set_it(client)
            client.extraction_pipelines.create(
                ExtractionPipeline(
                    external_id=self.external_id,
                    name=self.external_id,
                    data_set_id=self._data_set_id,
                    description=self.description,
                )
            )
        return self

    def create_pipeline_run(
        self, client: CogniteClient, is_dry_run: bool = False, error_logger: Optional[Callable[[str], None]] = None
    ) -> PipelineRun:
        """

        Args:
            client: An instance of the CogniteClient.
            is_dry_run: Whether to run the pipeline as a dry run. This will not create a pipeline run.
            error_logger: The function to use for logging errors. Defaults to print.

        Returns:
            Return a pipeline run that can be used as a context manager.
        """
        self._data_set_id = self._data_set_id or self._get_data_set_it(client)

        return PipelineRun(client, self.external_id, self.config, self._data_set_id, error_logger or print, is_dry_run)

    def _get_data_set_it(self, client: CogniteClient) -> int:
        data_set = client.data_sets.retrieve(external_id=self.dataset_external_id)
        if data_set is None:
            raise ValueError(f"Dataset {self.dataset_external_id} does not exist")
        if data_set.id is None:
            raise ValueError(f"Dataset {self.dataset_external_id} does not have an id")
        return data_set.id



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)

if TYPE_CHECKING:
    from ._bid_method import BidMethod, BidMethodGraphQL, BidMethodWrite
    from ._plant import Plant, PlantGraphQL, PlantWrite
    from ._watercourse import Watercourse, WatercourseGraphQL, WatercourseWrite


__all__ = [
    "PriceArea",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
]


PriceAreaTextFields = Literal[
    "name",
    "display_name",
    "description",
    "timezone",
    "capacity_price_up",
    "capacity_price_down",
    "activation_price_up",
    "activation_price_down",
    "relative_activation",
    "total_capacity_allocation_up",
    "total_capacity_allocation_down",
    "own_capacity_allocation_up",
    "own_capacity_allocation_down",
    "main_scenario_day_ahead",
    "day_ahead_price",
]
PriceAreaFields = Literal[
    "name",
    "display_name",
    "description",
    "timezone",
    "capacity_price_up",
    "capacity_price_down",
    "activation_price_up",
    "activation_price_down",
    "relative_activation",
    "total_capacity_allocation_up",
    "total_capacity_allocation_down",
    "own_capacity_allocation_up",
    "own_capacity_allocation_down",
    "main_scenario_day_ahead",
    "day_ahead_price",
]

_PRICEAREA_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "description": "description",
    "timezone": "timezone",
    "capacity_price_up": "capacityPriceUp",
    "capacity_price_down": "capacityPriceDown",
    "activation_price_up": "activationPriceUp",
    "activation_price_down": "activationPriceDown",
    "relative_activation": "relativeActivation",
    "total_capacity_allocation_up": "totalCapacityAllocationUp",
    "total_capacity_allocation_down": "totalCapacityAllocationDown",
    "own_capacity_allocation_up": "ownCapacityAllocationUp",
    "own_capacity_allocation_down": "ownCapacityAllocationDown",
    "main_scenario_day_ahead": "mainScenarioDayAhead",
    "day_ahead_price": "dayAheadPrice",
}


class PriceAreaGraphQL(GraphQLCore):
    """This represents the reading version of price area, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: Name for the PriceArea.
        display_name: Display name for the PriceArea.
        description: Description for the PriceArea.
        timezone: The timezone of the price area
        capacity_price_up: The capacity price up field.
        capacity_price_down: The capacity price down field.
        activation_price_up: The mFRR activation price (TBC)
        activation_price_down: The mFRR activate price (TBC)
        relative_activation: Value between -1 (100 % activation down) and 1 (100 % activation down)
        total_capacity_allocation_up: The total capacity allocation up field.
        total_capacity_allocation_down: The total capacity allocation down field.
        own_capacity_allocation_up: The own capacity allocation up field.
        own_capacity_allocation_down: The own capacity allocation down field.
        default_method_day_ahead: Default method for day ahead bids
        main_scenario_day_ahead: Main scenario for day ahead bids
        day_ahead_price: Day ahead price for the price area
        plants: The plants that are connected to the Watercourse.
        watercourses: The watercourses that are connected to the PriceArea.
    """

    view_id = dm.ViewId("power-ops-assets", "PriceArea", "1")
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    description: Optional[str] = None
    timezone: Optional[str] = None
    capacity_price_up: Union[TimeSeries, dict, None] = Field(None, alias="capacityPriceUp")
    capacity_price_down: Union[TimeSeries, dict, None] = Field(None, alias="capacityPriceDown")
    activation_price_up: Union[TimeSeries, dict, None] = Field(None, alias="activationPriceUp")
    activation_price_down: Union[TimeSeries, dict, None] = Field(None, alias="activationPriceDown")
    relative_activation: Union[TimeSeries, dict, None] = Field(None, alias="relativeActivation")
    total_capacity_allocation_up: Union[TimeSeries, dict, None] = Field(None, alias="totalCapacityAllocationUp")
    total_capacity_allocation_down: Union[TimeSeries, dict, None] = Field(None, alias="totalCapacityAllocationDown")
    own_capacity_allocation_up: Union[TimeSeries, dict, None] = Field(None, alias="ownCapacityAllocationUp")
    own_capacity_allocation_down: Union[TimeSeries, dict, None] = Field(None, alias="ownCapacityAllocationDown")
    default_method_day_ahead: Optional[BidMethodGraphQL] = Field(None, repr=False, alias="defaultMethodDayAhead")
    main_scenario_day_ahead: Union[TimeSeries, dict, None] = Field(None, alias="mainScenarioDayAhead")
    day_ahead_price: Union[TimeSeries, dict, None] = Field(None, alias="dayAheadPrice")
    plants: Optional[list[PlantGraphQL]] = Field(default=None, repr=False)
    watercourses: Optional[list[WatercourseGraphQL]] = Field(default=None, repr=False)

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("default_method_day_ahead", "plants", "watercourses", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PriceArea:
        """Convert this GraphQL format of price area to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return PriceArea(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            display_name=self.display_name,
            description=self.description,
            timezone=self.timezone,
            capacity_price_up=self.capacity_price_up,
            capacity_price_down=self.capacity_price_down,
            activation_price_up=self.activation_price_up,
            activation_price_down=self.activation_price_down,
            relative_activation=self.relative_activation,
            total_capacity_allocation_up=self.total_capacity_allocation_up,
            total_capacity_allocation_down=self.total_capacity_allocation_down,
            own_capacity_allocation_up=self.own_capacity_allocation_up,
            own_capacity_allocation_down=self.own_capacity_allocation_down,
            default_method_day_ahead=(
                self.default_method_day_ahead.as_read()
                if isinstance(self.default_method_day_ahead, GraphQLCore)
                else self.default_method_day_ahead
            ),
            main_scenario_day_ahead=self.main_scenario_day_ahead,
            day_ahead_price=self.day_ahead_price,
            plants=[plant.as_read() if isinstance(plant, GraphQLCore) else plant for plant in self.plants or []],
            watercourses=[
                watercourse.as_read() if isinstance(watercourse, GraphQLCore) else watercourse
                for watercourse in self.watercourses or []
            ],
        )

    def as_write(self) -> PriceAreaWrite:
        """Convert this GraphQL format of price area to the writing format."""
        return PriceAreaWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            display_name=self.display_name,
            description=self.description,
            timezone=self.timezone,
            capacity_price_up=self.capacity_price_up,
            capacity_price_down=self.capacity_price_down,
            activation_price_up=self.activation_price_up,
            activation_price_down=self.activation_price_down,
            relative_activation=self.relative_activation,
            total_capacity_allocation_up=self.total_capacity_allocation_up,
            total_capacity_allocation_down=self.total_capacity_allocation_down,
            own_capacity_allocation_up=self.own_capacity_allocation_up,
            own_capacity_allocation_down=self.own_capacity_allocation_down,
            default_method_day_ahead=(
                self.default_method_day_ahead.as_write()
                if isinstance(self.default_method_day_ahead, DomainModel)
                else self.default_method_day_ahead
            ),
            main_scenario_day_ahead=self.main_scenario_day_ahead,
            day_ahead_price=self.day_ahead_price,
            plants=[plant.as_write() if isinstance(plant, DomainModel) else plant for plant in self.plants or []],
            watercourses=[
                watercourse.as_write() if isinstance(watercourse, DomainModel) else watercourse
                for watercourse in self.watercourses or []
            ],
        )


class PriceArea(DomainModel):
    """This represents the reading version of price area.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: Name for the PriceArea.
        display_name: Display name for the PriceArea.
        description: Description for the PriceArea.
        timezone: The timezone of the price area
        capacity_price_up: The capacity price up field.
        capacity_price_down: The capacity price down field.
        activation_price_up: The mFRR activation price (TBC)
        activation_price_down: The mFRR activate price (TBC)
        relative_activation: Value between -1 (100 % activation down) and 1 (100 % activation down)
        total_capacity_allocation_up: The total capacity allocation up field.
        total_capacity_allocation_down: The total capacity allocation down field.
        own_capacity_allocation_up: The own capacity allocation up field.
        own_capacity_allocation_down: The own capacity allocation down field.
        default_method_day_ahead: Default method for day ahead bids
        main_scenario_day_ahead: Main scenario for day ahead bids
        day_ahead_price: Day ahead price for the price area
        plants: The plants that are connected to the Watercourse.
        watercourses: The watercourses that are connected to the PriceArea.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "PriceArea")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    description: Optional[str] = None
    timezone: str
    capacity_price_up: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceUp")
    capacity_price_down: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceDown")
    activation_price_up: Union[TimeSeries, str, None] = Field(None, alias="activationPriceUp")
    activation_price_down: Union[TimeSeries, str, None] = Field(None, alias="activationPriceDown")
    relative_activation: Union[TimeSeries, str, None] = Field(None, alias="relativeActivation")
    total_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationUp")
    total_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationDown")
    own_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationUp")
    own_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationDown")
    default_method_day_ahead: Union[BidMethod, str, dm.NodeId, None] = Field(
        None, repr=False, alias="defaultMethodDayAhead"
    )
    main_scenario_day_ahead: Union[TimeSeries, str, None] = Field(None, alias="mainScenarioDayAhead")
    day_ahead_price: Union[TimeSeries, str, None] = Field(None, alias="dayAheadPrice")
    plants: Union[list[Plant], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)
    watercourses: Union[list[Watercourse], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def as_write(self) -> PriceAreaWrite:
        """Convert this read version of price area to the writing version."""
        return PriceAreaWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            description=self.description,
            timezone=self.timezone,
            capacity_price_up=self.capacity_price_up,
            capacity_price_down=self.capacity_price_down,
            activation_price_up=self.activation_price_up,
            activation_price_down=self.activation_price_down,
            relative_activation=self.relative_activation,
            total_capacity_allocation_up=self.total_capacity_allocation_up,
            total_capacity_allocation_down=self.total_capacity_allocation_down,
            own_capacity_allocation_up=self.own_capacity_allocation_up,
            own_capacity_allocation_down=self.own_capacity_allocation_down,
            default_method_day_ahead=(
                self.default_method_day_ahead.as_write()
                if isinstance(self.default_method_day_ahead, DomainModel)
                else self.default_method_day_ahead
            ),
            main_scenario_day_ahead=self.main_scenario_day_ahead,
            day_ahead_price=self.day_ahead_price,
            plants=[plant.as_write() if isinstance(plant, DomainModel) else plant for plant in self.plants or []],
            watercourses=[
                watercourse.as_write() if isinstance(watercourse, DomainModel) else watercourse
                for watercourse in self.watercourses or []
            ],
        )

    def as_apply(self) -> PriceAreaWrite:
        """Convert this read version of price area to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaWrite(DomainModelWrite):
    """This represents the writing version of price area.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area.
        data_record: The data record of the price area node.
        name: Name for the PriceArea.
        display_name: Display name for the PriceArea.
        description: Description for the PriceArea.
        timezone: The timezone of the price area
        capacity_price_up: The capacity price up field.
        capacity_price_down: The capacity price down field.
        activation_price_up: The mFRR activation price (TBC)
        activation_price_down: The mFRR activate price (TBC)
        relative_activation: Value between -1 (100 % activation down) and 1 (100 % activation down)
        total_capacity_allocation_up: The total capacity allocation up field.
        total_capacity_allocation_down: The total capacity allocation down field.
        own_capacity_allocation_up: The own capacity allocation up field.
        own_capacity_allocation_down: The own capacity allocation down field.
        default_method_day_ahead: Default method for day ahead bids
        main_scenario_day_ahead: Main scenario for day ahead bids
        day_ahead_price: Day ahead price for the price area
        plants: The plants that are connected to the Watercourse.
        watercourses: The watercourses that are connected to the PriceArea.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power-ops-types", "PriceArea")
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    description: Optional[str] = None
    timezone: str
    capacity_price_up: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceUp")
    capacity_price_down: Union[TimeSeries, str, None] = Field(None, alias="capacityPriceDown")
    activation_price_up: Union[TimeSeries, str, None] = Field(None, alias="activationPriceUp")
    activation_price_down: Union[TimeSeries, str, None] = Field(None, alias="activationPriceDown")
    relative_activation: Union[TimeSeries, str, None] = Field(None, alias="relativeActivation")
    total_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationUp")
    total_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="totalCapacityAllocationDown")
    own_capacity_allocation_up: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationUp")
    own_capacity_allocation_down: Union[TimeSeries, str, None] = Field(None, alias="ownCapacityAllocationDown")
    default_method_day_ahead: Union[BidMethodWrite, str, dm.NodeId, None] = Field(
        None, repr=False, alias="defaultMethodDayAhead"
    )
    main_scenario_day_ahead: Union[TimeSeries, str, None] = Field(None, alias="mainScenarioDayAhead")
    day_ahead_price: Union[TimeSeries, str, None] = Field(None, alias="dayAheadPrice")
    plants: Union[list[PlantWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)
    watercourses: Union[list[WatercourseWrite], list[str], list[dm.NodeId], None] = Field(default=None, repr=False)

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(PriceArea, dm.ViewId("power-ops-assets", "PriceArea", "1"))

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.description is not None or write_none:
            properties["description"] = self.description

        if self.timezone is not None:
            properties["timezone"] = self.timezone

        if self.capacity_price_up is not None or write_none:
            if isinstance(self.capacity_price_up, str) or self.capacity_price_up is None:
                properties["capacityPriceUp"] = self.capacity_price_up
            else:
                properties["capacityPriceUp"] = self.capacity_price_up.external_id

        if self.capacity_price_down is not None or write_none:
            if isinstance(self.capacity_price_down, str) or self.capacity_price_down is None:
                properties["capacityPriceDown"] = self.capacity_price_down
            else:
                properties["capacityPriceDown"] = self.capacity_price_down.external_id

        if self.activation_price_up is not None or write_none:
            if isinstance(self.activation_price_up, str) or self.activation_price_up is None:
                properties["activationPriceUp"] = self.activation_price_up
            else:
                properties["activationPriceUp"] = self.activation_price_up.external_id

        if self.activation_price_down is not None or write_none:
            if isinstance(self.activation_price_down, str) or self.activation_price_down is None:
                properties["activationPriceDown"] = self.activation_price_down
            else:
                properties["activationPriceDown"] = self.activation_price_down.external_id

        if self.relative_activation is not None or write_none:
            if isinstance(self.relative_activation, str) or self.relative_activation is None:
                properties["relativeActivation"] = self.relative_activation
            else:
                properties["relativeActivation"] = self.relative_activation.external_id

        if self.total_capacity_allocation_up is not None or write_none:
            if isinstance(self.total_capacity_allocation_up, str) or self.total_capacity_allocation_up is None:
                properties["totalCapacityAllocationUp"] = self.total_capacity_allocation_up
            else:
                properties["totalCapacityAllocationUp"] = self.total_capacity_allocation_up.external_id

        if self.total_capacity_allocation_down is not None or write_none:
            if isinstance(self.total_capacity_allocation_down, str) or self.total_capacity_allocation_down is None:
                properties["totalCapacityAllocationDown"] = self.total_capacity_allocation_down
            else:
                properties["totalCapacityAllocationDown"] = self.total_capacity_allocation_down.external_id

        if self.own_capacity_allocation_up is not None or write_none:
            if isinstance(self.own_capacity_allocation_up, str) or self.own_capacity_allocation_up is None:
                properties["ownCapacityAllocationUp"] = self.own_capacity_allocation_up
            else:
                properties["ownCapacityAllocationUp"] = self.own_capacity_allocation_up.external_id

        if self.own_capacity_allocation_down is not None or write_none:
            if isinstance(self.own_capacity_allocation_down, str) or self.own_capacity_allocation_down is None:
                properties["ownCapacityAllocationDown"] = self.own_capacity_allocation_down
            else:
                properties["ownCapacityAllocationDown"] = self.own_capacity_allocation_down.external_id

        if self.default_method_day_ahead is not None:
            properties["defaultMethodDayAhead"] = {
                "space": (
                    self.space
                    if isinstance(self.default_method_day_ahead, str)
                    else self.default_method_day_ahead.space
                ),
                "externalId": (
                    self.default_method_day_ahead
                    if isinstance(self.default_method_day_ahead, str)
                    else self.default_method_day_ahead.external_id
                ),
            }

        if self.main_scenario_day_ahead is not None or write_none:
            if isinstance(self.main_scenario_day_ahead, str) or self.main_scenario_day_ahead is None:
                properties["mainScenarioDayAhead"] = self.main_scenario_day_ahead
            else:
                properties["mainScenarioDayAhead"] = self.main_scenario_day_ahead.external_id

        if self.day_ahead_price is not None or write_none:
            if isinstance(self.day_ahead_price, str) or self.day_ahead_price is None:
                properties["dayAheadPrice"] = self.day_ahead_price
            else:
                properties["dayAheadPrice"] = self.day_ahead_price.external_id

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "isSubAssetOf")
        for plant in self.plants or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=plant,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power-ops-types", "isSubAssetOf")
        for watercourse in self.watercourses or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=watercourse,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.default_method_day_ahead, DomainModelWrite):
            other_resources = self.default_method_day_ahead._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.capacity_price_up, CogniteTimeSeries):
            resources.time_series.append(self.capacity_price_up)

        if isinstance(self.capacity_price_down, CogniteTimeSeries):
            resources.time_series.append(self.capacity_price_down)

        if isinstance(self.activation_price_up, CogniteTimeSeries):
            resources.time_series.append(self.activation_price_up)

        if isinstance(self.activation_price_down, CogniteTimeSeries):
            resources.time_series.append(self.activation_price_down)

        if isinstance(self.relative_activation, CogniteTimeSeries):
            resources.time_series.append(self.relative_activation)

        if isinstance(self.total_capacity_allocation_up, CogniteTimeSeries):
            resources.time_series.append(self.total_capacity_allocation_up)

        if isinstance(self.total_capacity_allocation_down, CogniteTimeSeries):
            resources.time_series.append(self.total_capacity_allocation_down)

        if isinstance(self.own_capacity_allocation_up, CogniteTimeSeries):
            resources.time_series.append(self.own_capacity_allocation_up)

        if isinstance(self.own_capacity_allocation_down, CogniteTimeSeries):
            resources.time_series.append(self.own_capacity_allocation_down)

        if isinstance(self.main_scenario_day_ahead, CogniteTimeSeries):
            resources.time_series.append(self.main_scenario_day_ahead)

        if isinstance(self.day_ahead_price, CogniteTimeSeries):
            resources.time_series.append(self.day_ahead_price)

        return resources


class PriceAreaApply(PriceAreaWrite):
    def __new__(cls, *args, **kwargs) -> PriceAreaApply:
        warnings.warn(
            "PriceAreaApply is deprecated and will be removed in v1.0. Use PriceAreaWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceArea.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PriceAreaList(DomainModelList[PriceArea]):
    """List of price areas in the read version."""

    _INSTANCE = PriceArea

    def as_write(self) -> PriceAreaWriteList:
        """Convert these read versions of price area to the writing versions."""
        return PriceAreaWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceAreaWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaWriteList(DomainModelWriteList[PriceAreaWrite]):
    """List of price areas in the writing version."""

    _INSTANCE = PriceAreaWrite


class PriceAreaApplyList(PriceAreaWriteList): ...


def _create_price_area_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    description: str | list[str] | None = None,
    description_prefix: str | None = None,
    timezone: str | list[str] | None = None,
    timezone_prefix: str | None = None,
    default_method_day_ahead: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if isinstance(description, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("description"), value=description))
    if description and isinstance(description, list):
        filters.append(dm.filters.In(view_id.as_property_ref("description"), values=description))
    if description_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("description"), value=description_prefix))
    if isinstance(timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timezone"), value=timezone))
    if timezone and isinstance(timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timezone"), values=timezone))
    if timezone_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timezone"), value=timezone_prefix))
    if default_method_day_ahead and isinstance(default_method_day_ahead, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultMethodDayAhead"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": default_method_day_ahead},
            )
        )
    if default_method_day_ahead and isinstance(default_method_day_ahead, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultMethodDayAhead"),
                value={"space": default_method_day_ahead[0], "externalId": default_method_day_ahead[1]},
            )
        )
    if (
        default_method_day_ahead
        and isinstance(default_method_day_ahead, list)
        and isinstance(default_method_day_ahead[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultMethodDayAhead"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in default_method_day_ahead],
            )
        )
    if (
        default_method_day_ahead
        and isinstance(default_method_day_ahead, list)
        and isinstance(default_method_day_ahead[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultMethodDayAhead"),
                values=[{"space": item[0], "externalId": item[1]} for item in default_method_day_ahead],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import abc
from abc import ABC
from collections.abc import Iterable
from typing import Any, Callable, ClassVar, Literal, TypeVar

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadataUpdate, SequenceUpdate, TimeSeries, TimeSeriesList
from pydantic import BaseModel
from typing_extensions import Self

from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base.resource_type import ResourceType, _T_Type
from cognite.powerops.utils.serialization import remove_read_only_fields

from .cdf_resources import CDFFile, CDFResource, CDFSequence
from .helpers import isinstance_list


class Model(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        sequences = [sequence for item in self._resource_types() for sequence in item.sequences()]
        sequences.extend(self._fields_of_type(CDFSequence))
        return sequences

    def files(self) -> list[CDFFile]:
        files = [file for item in self._resource_types() for file in item.files()]
        files.extend(self._fields_of_type(CDFFile))
        return files

    def timeseries(self) -> TimeSeriesList:
        time_series = [ts for item in self._resource_types() for ts in item.time_series()]
        time_series.extend(self._fields_of_type(TimeSeries))
        return TimeSeriesList(time_series)

    cdf_resources: ClassVar[dict[Callable, type]] = {sequences: CDFSequence, files: CDFFile, timeseries: TimeSeries}

    def _resource_types(self) -> Iterable[ResourceType]:
        yield from self._fields_of_type(ResourceType)

    def _fields_of_type(self, type_: type[_T_Type]) -> Iterable[_T_Type]:
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if isinstance(value, type_):
                yield value
            elif isinstance_list(value, type_):
                yield from value

    @property
    def model_name(self) -> str:
        return type(self).__name__

    @classmethod
    def name(cls) -> str:
        return cls.__name__

    def dump_as_cdf_resource(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for resource_fun in self.cdf_resources.keys():
            name = resource_fun.__name__
            if items := resource_fun(self):

                def dump(resource: Any) -> dict[str, Any]:
                    return remove_read_only_fields(resource.dump(camel_case=True))

                output[name] = [dump(item) for item in items]
        return output

    @classmethod
    def _load_by_type_external_id(cls, data: dict[str, Any]) -> dict[str, dict[str, Any]]:
        loaded_by_type_external_id: dict[str, dict[str, Any]] = {}
        for function, resource_cls in cls.cdf_resources.items():
            name = function.__name__
            if items := data.get(name):
                loaded_by_type_external_id[name] = {
                    loaded.external_id: loaded
                    for loaded in (resource_cls._load(item) if isinstance(item, dict) else item for item in items)
                }
        return loaded_by_type_external_id

    @classmethod
    @abc.abstractmethod
    def load_from_cdf_resources(
        cls: type[Self], data: dict[str, Any], link: Literal["external_id", "object"] = "object"
    ) -> Self:
        """
        This function loads the model from the data retrieved from CDF.

        Args:
            data: The data retrieved from CDF.
            link: (Only data models) Whether to link the data model using external IDs or objects. Linking
                objects directly can lead to circular dependencies, thus if this is a concern you should link
                by "external_id".

        Returns:
            Instance of the model.
        """
        raise NotImplementedError()

    @classmethod
    def _external_id_key(cls, resource: dict | Any) -> str:
        if hasattr(resource, "external_id"):
            return resource.external_id.casefold()
        elif isinstance(resource, dict) and ("external_id" in resource or "externalId" in resource):
            return resource.get("external_id", resource.get("externalId")).casefold()
        raise ValueError(f"Could not find external_id in {resource}")

    @classmethod
    @abc.abstractmethod
    def from_cdf(cls: type[T_Model], client: PowerOpsClient, data_set_external_id: str) -> T_Model: ...

    @abc.abstractmethod
    def standardize(self) -> None:
        """
        This ensures that the model is in a standardized form.

        This is useful when doing comparisons between models; as for example, the ordering of the assets in some
        lists does not matter.
        """
        raise NotImplementedError()

    @classmethod
    def _add_missing_hash(cls, client: CogniteClient, resource: list[CDFResource]):
        """
        This function adds the hash to the metadata of the resource if it is missing.
        The intention is for backwards compatibility with resources that were created before the hash was added.
        """
        for item in resource:
            resource = item.cdf_resource
            if not resource.metadata:
                resource.metadata = {}
            if CDFResource.content_key_hash not in resource.metadata:
                if isinstance(item, CDFSequence):
                    content = client.sequences.data.retrieve_dataframe(
                        start=0, end=-1, external_id=item.external_id, limit=None
                    )
                    resource.metadata[CDFResource.content_key_hash] = CDFSequence.calculate_hash(content)
                    update = SequenceUpdate(external_id=item.external_id).metadata.add(
                        {CDFSequence.content_key_hash: resource.metadata[CDFResource.content_key_hash]}
                    )
                    client.sequences.update(update)
                elif isinstance(item, CDFFile):
                    content = client.files.download_bytes(item.external_id)
                    resource.metadata[CDFResource.content_key_hash] = CDFFile.calculate_hash(content)
                    update = FileMetadataUpdate(external_id=item.external_id).metadata.add(
                        {CDFFile.content_key_hash: resource.metadata[CDFResource.content_key_hash]}
                    )
                    client.files.update(update)
                else:
                    raise NotImplementedError(f"Cannot get content for {item}")


T_Model = TypeVar("T_Model", bound=Model)



from dataclasses import dataclass
from pathlib import Path

from cognite.client.data_classes.data_modeling import DataModelId


@dataclass
class PowerOpsGraphQLModel:
    name: str
    description: str
    graphql_file: Path
    id_: DataModelId
    extra_types: list[Path] = None

    @property
    def graphql(self) -> str:
        graphql = self.graphql_file.read_text()
        if self.extra_types:
            graphql = "\n\n".join([extra.read_text() for extra in self.extra_types] + [graphql])

        return graphql

    def spaces(self) -> list[str]:
        return [self.id_.space]



from __future__ import annotations

import json
from collections.abc import Sequence
from typing import Optional

from cognite.client.data_classes import Asset, Label, Relationship
from pydantic import BaseModel, field_validator

from cognite.powerops.cdf_labels import RelationshipLabel
from cognite.powerops.resync.config._shared import ExternalId

p_min_fallback = 0
p_max_fallback = 1e20

head_loss_factor_fallback = 0.0

PlantName = str


class Plant(BaseModel):
    name: PlantName
    external_id: ExternalId
    outlet_level: float  # meters above sea level
    p_min: float = p_min_fallback
    p_max: float = p_max_fallback  # arbitrary max power output if not specified
    head_loss_factor: float = 0.0
    penstock_head_loss_factors: Optional[dict[str, float]] = None
    connections_losses: Optional[float] = None
    display_name: Optional[str] = None
    ordering_key: Optional[float] = None

    generator_ext_ids: list[ExternalId] = []  # external IDs of generator assets
    inlet_reservoir_ext_id: Optional[ExternalId] = None  # external ID of reservoir asset

    inlet_level_time_series: Optional[ExternalId] = None  # external ID of time series with values in m.a.s.l.
    outlet_level_time_series: Optional[ExternalId] = None  # external ID of time series with values in m.a.s.l.
    water_value_time_series: Optional[ExternalId] = None  # external ID of time series with values in €/MWh
    feeding_fee_time_series: Optional[ExternalId] = None  # external ID of time series with values in percent
    p_min_time_series: Optional[ExternalId] = None  # external ID of time series with values in MW
    p_max_time_series: Optional[ExternalId] = None  # external ID of time series with values in MW
    head_direct_time_series: Optional[ExternalId] = None  # external ID of time series with values in m

    @field_validator("penstock_head_loss_factors", mode="before")
    @classmethod
    def parse_dict(cls, value):
        if isinstance(value, str):
            value = json.loads(value)
        return value

    @classmethod
    def from_cdf_resources(cls, asset: Asset, relationships: list[Relationship], **kwargs) -> Plant:
        """Initialise a Plant from CDF Asset and Relationships

        Args:
            asset: The plant Asset relationships
            relationships: Relationships to related resources (will be mapped to attributes based on labels)
            kwargs: Any other attributes that are not part of the Asset

        """
        # Initialise plant from Asset
        plant = cls(
            name=asset.name,
            external_id=asset.external_id,
            outlet_level=float(asset.metadata["outlet_level"]),
            p_min=float(asset.metadata["p_min"]),
            p_max=float(asset.metadata["p_max"]),
            head_loss_factor=float(asset.metadata["head_loss_factor"]),
            connection_losses=asset.metadata.get("connection_losses") or None,
            penstock_head_loss_factors=json.loads(asset.metadata.get("penstock_head_loss_factors") or "{}"),
            **kwargs,  # kwargs to set any other attributes that are not part of the Asset
        )

        # Find time series based on relationships
        time_series_ext_ids_and_labels = [
            (rel.target_external_id, [label.external_id for label in (rel.labels or [])])
            for rel in relationships
            if rel.target_type == "TIMESERIES" and rel.source_external_id == plant.external_id
        ]
        for ts_ext_id, labels in time_series_ext_ids_and_labels:
            if RelationshipLabel.INLET_LEVEL_TIME_SERIES in labels:
                plant.inlet_level_time_series = ts_ext_id
            elif RelationshipLabel.OUTLET_LEVEL_TIME_SERIES in labels:
                plant.outlet_level_time_series = ts_ext_id
            elif RelationshipLabel.WATER_VALUE_TIME_SERIES in labels:
                plant.water_value_time_series = ts_ext_id
            elif RelationshipLabel.FEEDING_FEE_TIME_SERIES in labels:
                plant.feeding_fee_time_series = ts_ext_id
            elif RelationshipLabel.P_MIN_TIME_SERIES in labels:
                plant.p_min_time_series = ts_ext_id
            elif RelationshipLabel.P_MAX_TIME_SERIES in labels:
                plant.p_max_time_series = ts_ext_id
            elif RelationshipLabel.HEAD_DIRECT_TIME_SERIES in labels:
                plant.head_direct_time_series = ts_ext_id

        # Find generators based on relationships
        plant.generator_ext_ids = [
            rel.target_external_id
            for rel in relationships
            if label_in_labels(RelationshipLabel.GENERATOR, rel.labels or [])
        ]

        # Find inlet reservoir based on relationships
        for rel in relationships:
            if (
                rel.target_type == "ASSET"
                and label_in_labels(RelationshipLabel.INLET_RESERVOIR, rel.labels or [])
                and rel.source_external_id == plant.external_id
            ):
                plant.inlet_reservoir_ext_id = rel.target_external_id
                break

        return plant


def label_in_labels(label_external_id: str, labels: Sequence[Label]) -> bool:
    """Check if a label with a given external id is in a list of labels."""
    return label_external_id in [label.external_id for label in labels]


class PlantTimeSeriesMapping(BaseModel):
    plant_name: str
    water_value: Optional[ExternalId] = None
    inlet_reservoir_level: Optional[ExternalId] = None
    outlet_reservoir_level: Optional[ExternalId] = None
    p_min: Optional[ExternalId] = None
    p_max: Optional[ExternalId] = None
    feeding_fee: Optional[ExternalId] = None
    head_direct: Optional[ExternalId] = None

    @field_validator("*", mode="before")
    @classmethod
    def parse_number_to_string(cls, value):
        return str(value) if isinstance(value, (int, float)) else value



from __future__ import annotations

from collections import Counter, defaultdict
from pathlib import Path
from typing import Any, Literal, Optional, overload

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from pydantic_core.core_schema import FieldValidationInfo
from typing_extensions import Self

from cognite.powerops.utils.serialization import load_yaml

from ._settings import Settings
from ._shared import TimeSeriesMapping, TimeSeriesMappingV2
from .cogshop.shop_file_config import ShopFileConfig
from .market import (
    BenchmarkingConfig,
    BidMatrixGeneratorConfig,
    BidProcessConfig,
    Market,
    PriceScenario,
    PriceScenarioV2,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    RkomMarketConfig,
)
from .production import GeneratorTimeSeriesMapping, PlantTimeSeriesMapping, WatercourseConfig


class Config(BaseModel):
    @classmethod
    def load_yamls(cls, config_dir_path: Path) -> dict[str, Any]:
        """
        Loads all yaml files in the config directory and returns a dictionary with the file name as key and the
        content as value.

        Note the config_dir_path can be nested, all yaml files in the directory and subdirectories will be loaded.

        Parameters
        ----------
        config_dir_path: Path
            The path to the config directory

        Returns
        -------
            Dictionary with all yaml files loaded.
        """
        config: dict[str, Any] = {}
        field_names = set(cls.model_fields)
        for file_path in config_dir_path.glob("**/*.yaml"):
            if file_path.stem not in field_names:
                continue
            config[file_path.stem] = load_yaml(file_path)
        return config


class MarketConfig(Config):
    market: Market
    benchmarks: list[BenchmarkingConfig]
    price_scenario_by_id: dict[str, PriceScenario]
    price_scenario_by_id_v2: dict[str, PriceScenarioV2]

    bidprocess: list[BidProcessConfig]
    bidmatrix_generators: list[BidMatrixGeneratorConfig]

    rkom_bid_process: list[RKOMBidProcessConfig]
    rkom_bid_combination: Optional[list[RKOMBidCombinationConfig]] = None
    rkom_market: RkomMarketConfig = Field(default_factory=RkomMarketConfig.default)

    @field_validator("price_scenario_by_id")
    @classmethod
    def no_duplicated_scenarios(cls, value: dict[str, PriceScenario]):
        scenarios_by_hash = defaultdict(list)
        for id_, s in value.items():
            scenarios_by_hash[hash(s.model_dump_json(exclude={"name"}))].append((id_, s))
        if duplicated := [s for s in scenarios_by_hash.values() if len(s) > 1]:
            sep = ") |\n\t\t\t ("
            raise ValueError(
                f"Scenarios  "
                f"({sep.join([', '.join([id_ for id_, _ in duplicate_set]) for duplicate_set in duplicated])}) "
                f"\nare duplicated."
            )
        return value

    @field_validator("rkom_bid_combination", mode="after")
    @classmethod
    def valid_process_external_id(cls, value: list[RKOMBidCombinationConfig], values: FieldValidationInfo):
        valid_ids = {process_config.external_id for process_config in values.data["rkom_bid_process"]}
        for combination in value:
            for external_id_to_validate in combination.rkom_bid_config_external_ids:
                if external_id_to_validate not in valid_ids:
                    raise ValueError(
                        f"Reference to rkom bid process config in rkom_bid_combination yaml is wrong for "
                        f"{external_id_to_validate}. "
                        f"Possible references are: {[config.external_id for config in values.data['rkom_bid_process']]}"
                    )
        return value

    @classmethod
    def instantiate_from_dict(cls, config: dict) -> MarketConfig:
        price_scenario_by_id_v2 = {
            k: PriceScenarioV2.load_from_dict(v) for k, v in config["price_scenario_by_id_v2"].items()
        }
        config["price_scenario_by_id_v2"] = price_scenario_by_id_v2
        return cls(**config)

    @field_validator("bidprocess", mode="after")
    @classmethod
    def one_default_per_price_area(cls, value: list[BidProcessConfig]):
        default_configs = Counter(
            bid_config.price_area_name for bid_config in value if bid_config.is_default_config_for_price_area
        )

        if price_areas_with_multiple_default_configs := [
            price_area for price_area, count in default_configs.items() if count > 1
        ]:
            raise ValueError(
                f"Multiple default bid configs for price areas: {price_areas_with_multiple_default_configs}"
            )
        return value


class ProductionConfig(Config):
    dayahead_price_timeseries: dict[str, str]
    watercourses: list[WatercourseConfig]
    generator_time_series_mappings: Optional[list[GeneratorTimeSeriesMapping]] = None
    plant_time_series_mappings: Optional[list[PlantTimeSeriesMapping]] = None

    @classmethod
    @overload
    def load_yamls(cls, config_dir_path: Path, instantiate: Literal[False] = False) -> dict[str, Any]: ...

    @classmethod
    @overload
    def load_yamls(cls, config_dir_path: Path, instantiate: Literal[True]) -> ProductionConfig: ...

    @classmethod
    def load_yamls(cls, config_dir_path: Path, instantiate: bool = False) -> dict[str, Any] | ProductionConfig:
        config: dict[str, Any] = {}
        for field_name in cls.model_fields:
            if not (config_file_path := config_dir_path / f"{field_name}.yaml").exists():
                continue
            content = load_yaml(config_file_path)
            if field_name == "watercourses":
                for watercourse in content:
                    # Complete the paths for the yaml files
                    if all(key in watercourse for key in ["directory", "model_raw"]):
                        watercourse["yaml_raw_path"] = (
                            config_dir_path / watercourse["directory"] / watercourse["model_raw"]
                        )
                        shop_model_template = load_yaml(watercourse["yaml_raw_path"], encoding="utf-8")
                        watercourse["shop_model_template"] = shop_model_template
                        watercourse["generators"] = _read_generators_from_shop_model_template(shop_model_template)
                    if all(key in watercourse for key in ["directory", "model_processed"]):
                        watercourse["yaml_processed_path"] = (
                            config_dir_path / watercourse["directory"] / watercourse["model_processed"]
                        )
            config[field_name] = content
        if instantiate:
            return cls(**config)
        return config


class CogShopConfig(Config):
    time_series_mappings: Optional[list[TimeSeriesMapping]] = None
    time_series_mappings_v2: Optional[list[TimeSeriesMappingV2]] = None
    watercourses_shop: list[ShopFileConfig]  # validation needs to happen here. If any of the files
    _dependent_shop_files: set = {
        "extra_data",
        "water_value_cut_file_reservoir_mapping",
        "water_value_cut_file",
        "module_series",
    }

    @field_validator("watercourses_shop")
    @classmethod
    def validate_shop_related_files(cls, value):
        """
        If user has added any of the files above to the configuration, then the "cog_shop_files_config" file
        must exist among the files to be uploaded
        """
        files_to_check = cls._dependent_shop_files.default.intersection({file.cogshop_file_type for file in value})
        if not files_to_check:
            return value
        extra_shop_files = [
            file.file_path.stem for file in value if file.cogshop_file_type in cls._dependent_shop_files.default
        ]
        has_cog_shop_files_config = "cog_shop_files_config" in extra_shop_files
        if has_cog_shop_files_config:
            extra_shop_files.remove("cog_shop_files_config")
            if not extra_shop_files:
                raise ValueError(
                    "Ensure that extra shop files to accompany cog_shop_files_config is added to watercourse"
                )
        elif extra_shop_files:
            raise ValueError(
                "Missing 'cog_shop_files_config.yaml'. "
                "This is needed for CogSHOP to know which order and type to load extra SHOP files."
            )
        return value

    @classmethod
    def instantiate_from_dict(cls, config: dict) -> CogShopConfig:
        time_series_mappings_v2 = [
            TimeSeriesMappingV2.load_from_dict(time_series_mapping)
            for time_series_mapping in config["time_series_mappings_v2"]
        ]
        config["time_series_mappings_v2"] = time_series_mappings_v2
        return cls(**config)


class ReSyncConfig(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    settings: Settings = Field(alias="constants")
    production: ProductionConfig
    market: MarketConfig
    cogshop: CogShopConfig

    @classmethod
    def from_yamls(cls, config_dir_path: Path, cdf_project: str) -> ReSyncConfig:
        configs: dict[str, Any] = {}
        for field in cls.model_fields:
            if (config_file_path := config_dir_path / f"{field}.yaml").exists():
                configs[field] = load_yaml(config_file_path, encoding="utf-8")
            elif (
                (config_subdir_path := config_dir_path / field).exists()
                and (class_ := cls.model_fields[field].annotation) is not None
                and issubclass(class_, Config)
            ):
                configs[field] = class_.load_yamls(config_subdir_path)

        watercourse_directory_by_name = {}
        for watercourse in configs.get("production", {}).get("watercourses", []):
            if "directory" in watercourse and "name" in watercourse:
                watercourse_directory_by_name[watercourse["name"]] = watercourse["directory"]

        # Completing the paths for the shop files
        for shop_file in configs.get("cogshop", {}).get("watercourses_shop", []):
            if (watercourse_name := shop_file.get("watercourse_name")) and (
                directory := watercourse_directory_by_name.get(watercourse_name)
            ):
                if "file_path" in shop_file:
                    shop_file["file_path"] = config_dir_path / "production" / directory / shop_file["file_path"]
                # For backwards compatibility
                if "path" in shop_file:
                    shop_file["path"] = config_dir_path / "production" / directory / shop_file["path"]

        # Todo Hack to get cdf project into settings.
        if "settings" in configs:
            configs["settings"]["cdf_project"] = cdf_project
        elif "constants" in configs:
            # For backwards compatibility
            configs["constants"]["cdf_project"] = cdf_project

        # Hack to instantiate CogShopConfig from dict
        for field_name in ["cogshop", "market"]:
            class_ = cls.model_fields[field_name].annotation
            configs[field_name] = class_.instantiate_from_dict(configs[field_name])

        return cls(**configs)

    @model_validator(mode="after")
    def object_exists_in_shop_template_model(self) -> Self:
        production = self.production
        time_series_mappings = self.cogshop.time_series_mappings or []
        # TODO Fix the assumption that timeseries mappings and watercourses are in the same order
        invalid_mappings_by_watercourse: dict[str, list[str]] = defaultdict(list)
        seen: set[tuple[str, str]] = set()
        for watercourse, timeseries_mapping in zip(production.watercourses, time_series_mappings):
            for mapping in timeseries_mapping:
                entry = (mapping.object_type.lower(), mapping.object_name.lower())
                if entry not in watercourse.valid_shop_objects and entry not in seen:
                    invalid_mappings_by_watercourse[watercourse.name].append(
                        f"{mapping.object_type}.{mapping.object_name}"
                    )
                    seen.add(entry)
        if invalid_mappings_by_watercourse:
            raise ValueError(
                "Timeseries mapping for watercourses not found in respective SHOP files: "
                f"{dict(invalid_mappings_by_watercourse)}"
            )

        return self


def _read_generators_from_shop_model_template(shop_model_template: dict[str, Any]) -> list[dict[str, Any]]:
    generators: list[dict[str, Any]] = []
    for generator_name, generator_attributes in shop_model_template.get("model", {}).get("generator", {}).items():
        attributes = generator_attributes.copy()
        attributes["name"] = generator_name
        generators.append(attributes)
    return generators



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "CommandsConfig",
    "CommandsConfigApply",
    "CommandsConfigList",
    "CommandsConfigApplyList",
    "CommandsConfigFields",
    "CommandsConfigTextFields",
]


CommandsConfigTextFields = Literal["commands"]
CommandsConfigFields = Literal["commands"]

_COMMANDSCONFIG_PROPERTIES_BY_FIELD = {
    "commands": "commands",
}


class CommandsConfig(DomainModel):
    space: str = "cogShop"
    commands: Optional[list[str]] = None

    def as_apply(self) -> CommandsConfigApply:
        return CommandsConfigApply(
            external_id=self.external_id,
            commands=self.commands,
        )


class CommandsConfigApply(DomainModelApply):
    space: str = "cogShop"
    commands: list[str]

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.commands is not None:
            properties["commands"] = self.commands
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("cogShop", "CommandsConfig"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class CommandsConfigList(TypeList[CommandsConfig]):
    _NODE = CommandsConfig

    def as_apply(self) -> CommandsConfigApplyList:
        return CommandsConfigApplyList([node.as_apply() for node in self.data])


class CommandsConfigApplyList(TypeApplyList[CommandsConfigApply]):
    _NODE = CommandsConfigApply



from __future__ import annotations

from typing import ClassVar, Optional, Union

from pydantic import Field, field_validator

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.base import CDFSequence
from cognite.powerops.utils.serialization import try_load_dict

from .base import Bid, Market, Process, ShopTransformation


class DayAheadBid(Bid):
    is_default_config_for_price_area: bool = True
    main_scenario: str
    price_area: str
    price_scenarios: dict[str, str]
    no_shop: bool = False
    bid_process_configuration_name: str
    bid_matrix_generator_config_external_id: str
    market_config_external_id: str

    @field_validator("price_scenarios", mode="before")
    @classmethod
    def parse_str(cls, value) -> dict:
        value = try_load_dict(value)
        if isinstance(value, list) and not value:
            return {}
        return value

    def standardize(self): ...


class DayAheadProcess(Process):
    parent_external_id = "bid_process_configurations"
    label: ClassVar[Union[AssetLabel, str]] = AssetLabel.BID_PROCESS_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations used in bid matrix generation processes"
    shop: ShopTransformation
    bid: DayAheadBid
    bid_matrix_generator_config: Optional[CDFSequence] = None
    incremental_mapping: list[CDFSequence] = Field(default_factory=list)

    @field_validator("incremental_mapping", mode="after")
    @classmethod
    def ordering(cls, value: list[CDFSequence]) -> list[CDFSequence]:
        return sorted(value, key=lambda x: x.external_id)

    @field_validator("bid", mode="before")
    @classmethod
    def parse_str(cls, value) -> dict:
        return try_load_dict(value)

    @property
    def watercourses(self) -> list[str]:
        return list({i.sequence.metadata.get("shop:watercourse") for i in self.incremental_mapping})

    def standardize(self) -> None:
        self.bid.standardize()
        self.incremental_mapping = self.ordering(self.incremental_mapping)


class NordPoolMarket(Market):
    # Temporary optional to fix that this is missing in CDF.
    max_price: Optional[float] = None
    min_price: Optional[float] = None
    price_steps: Optional[int] = None
    price_unit: Optional[str] = None
    tick_size: Optional[float] = None
    time_unit: Optional[str] = None
    trade_lot: Optional[float] = None
